# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *              J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module implement some base classes and utils for wizards
The content of this module is not discovered at runtime by pyworkflow.
Usage of this content is though importing it
"""

import tkinter as tk
import tkinter.ttk as ttk

import pyworkflow as pw
import pyworkflow.object as pwobj
import pyworkflow.wizard as pwizard
import pyworkflow.gui.dialog as dialog
from matplotlib import cm
from pwem import convertPixToLength, splitRange
from pyworkflow.gui.plotter import getHexColorList
from pyworkflow.gui.tree import BoundTree, ListTreeProvider, ListTreeProviderString, AttributesTreeProvider
from pyworkflow.gui.widgets import LabelSlider, ExplanationText

import pwem.constants as emcts
import pwem.objects as emobj
from pwem import emlib
from pyworkflow.protocol import IntParam, StringParam, FloatParam, LEVEL_ADVANCED

# Color map wizard constants
HIGHEST_ATTR = 'highest'
LOWEST_ATTR = 'lowest'
INTERVAL_ATTR = 'intervals'
COLORMAP_ATTR = 'colorMap'


# ===============================================================================
#    Wizard EM base class
# ===============================================================================

class EmWizard(pwizard.Wizard):

    def _getMics(self, objs):
        return [mic.clone() for mic in objs]

    def _getParticles(self, objs, num=100):
        particles = []
        for i, par in enumerate(objs):
            # Cloning the particle
            particle = par.clone()
            if i == num:  # Only load up to NUM particles
                break
            particles.append(particle)

        return particles

    def _getVols(self, objs):
        vols = []
        if isinstance(objs, emobj.Volume):
            vols.append(objs)
        else:
            vols = [vol.clone() for vol in objs]

        return vols

    def _getListProvider(self, objs):
        """ This should be implemented to return the list
        of object to be displayed in the tree.
        """
        provider = None
        if objs.hasValue():
            # If objs is a PointerList currently it can only be formed of SetOfVolumes and Volume
            # (for protocol align_volume). Should this change review this part
            if isinstance(objs, pwobj.PointerList):
                vols_total = []
                for pointer in objs:
                    obj = pointer.get()
                    vols = self._getVols(obj)
                    vols_total.extend(vols)
                provider = ListTreeProvider(vols_total)
            else:
                objs = objs.get()

                if isinstance(objs, emobj.SetOfMicrographs):
                    mics = self._getMics(objs)
                    provider = ListTreeProvider(mics)

                if isinstance(objs, emobj.SetOfParticles):
                    particles = self._getParticles(objs)
                    provider = ListTreeProvider(particles)

                if isinstance(objs, emobj.SetOfVolumes) or isinstance(objs,
                                                                      emobj.Volume):
                    vols = self._getVols(objs)
                    provider = ListTreeProvider(vols)

            return provider
        return None

    def _getInputProtocol(self, targets, protocol):
        label = []
        value = []

        for k, v in targets:
            if k.__name__ == protocol.getClassName():
                label = v
                for val in v:
                    value.append(protocol.getAttributeValue(val))

        if len(label) > 1:
            return label, value
        else:
            return label[0], value[0]


# ===============================================================================
#    Wizards base classes
# ===============================================================================

class VariableWizard(pwizard.Wizard):
    """Wizard base class object where input and output paramNames can be modified and added, so
    one wizard can be used in several protocols with parameters of different names"""
    #Variables _targets, _inputs and _outputs must be created in sons
    # _targets, _inputs, _outputs = [], {}, {}

    def addTarget(self, protocol, targets, inputs, outputs):
        '''Add a target to a wizard and the input and output parameters are stored in a dictionary with
        (protocol, targetParamName) as key.
        Targets must be added one by one.'''
        self._targets += [(protocol, targets)]
        self._inputs.update({(protocol, targets[0]): inputs})
        self._outputs.update({(protocol, targets[0]): outputs})

    def getInputOutput(self, form):
        '''Retrieving input and output paramNames corresponding to the protocol and target of the wizard clicked'''
        outParam = ''
        for target in self._targets:
            if form.wizParamName == target[1][0] and form.protocol.__class__ == target[0]:
                prot, target = (target[0], target[1][0])
                inParams, outParam = self._inputs[(prot, target)], self._outputs[(prot, target)]
                inParams = self.filterPresentInputs(inParams, form.protocol)
        return inParams, outParam

    def filterPresentInputs(self, inputParams, protocol):
        '''Filter the inputs to allow flexibility:
        1) InputParams can be lists: the first not None parameter of the list will be chosen.
        2) InputParams can be dict: key -> EnumParam (or IntParam) which functions as index
                                    value -> list containing paramNames. The hey index element will be chosen
        '''
        fInpParams = []
        for inPar in inputParams:
            if type(inPar) == list:
                present = ''
                for option in inPar:
                    if getattr(protocol, option).get():
                        present = option
                        break
                fInpParams.append(present)
            elif type(inPar) == dict:
                idxParam = list(inPar.keys())[0]
                optionParams = inPar[idxParam]

                idx = getattr(protocol, idxParam).get()
                fInpParams.append(optionParams[idx])
            else:
                fInpParams.append(inPar)
        return fInpParams

class DownsampleWizard(EmWizard):

    def show(self, form, value, label, units=emcts.UNIT_PIXEL):
        protocol = form.protocol
        provider = self._getProvider(protocol)

        if provider is not None:
            args = {'unit': units,
                    'downsample': value
                    }

            d = DownsampleDialog(form.root, provider, **args)
            if d.resultYes():
                form.setVar(label, d.getDownsample())
        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)

    @classmethod
    def getView(self):
        return "wiz_downsampling"


class CtfWizard(EmWizard):

    def show(self, form, value, label, units=emcts.UNIT_PIXEL):
        protocol = form.protocol
        #        form.setParamFromVar('inputMicrographs') # update selected input micrographs
        provider = self._getProvider(protocol)

        if provider is not None:
            args = {'unit': units,
                    'lf': value[0],
                    'hf': value[1]
                    }
            d = CtfDialog(form.root, provider, **args)

            if d.resultYes():
                form.setVar(label[0], d.getLowFreq())
                form.setVar(label[1], d.getHighFreq())
        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)

    @classmethod
    def getView(self):
        return "wiz_ctf"


class MaskRadiusWizard(EmWizard):

    def show(self, form, value, label, units=emcts.UNIT_PIXEL):
        protocol = form.protocol
        provider = self._getProvider(protocol)

        if provider is not None:
            d = MaskPreviewDialog(form.root,
                                  provider,
                                  maskRadius=value,
                                  unit=units)
            if d.resultYes():
                if units == emcts.UNIT_ANGSTROM:
                    value = round(d.getRadiusAngstroms(d.radiusSlider))  # Must be an integer
                else:
                    value = d.getRadius(d.radiusSlider)
                self.setVar(form, label, value)
        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)

    def setVar(self, form, label, value):
        form.setVar(label, value)


class MaskRadiiWizard(EmWizard):

    def show(self, form, value, label, units=emcts.UNIT_PIXEL):
        protocol = form.protocol
        provider = self._getProvider(protocol)

        if provider is not None:
            d = MaskRadiiPreviewDialog(form.root,
                                       provider,
                                       innerRadius=value[0],
                                       outerRadius=value[1],
                                       unit=units)
            if d.resultYes():
                if units == emcts.UNIT_ANGSTROM:
                    form.setVar(label[0], d.getRadiusAngstroms(d.radiusSliderIn))
                    form.setVar(label[1], d.getRadiusAngstroms(d.radiusSliderOut))
                else:
                    form.setVar(label[0], d.getRadius(d.radiusSliderIn))
                    form.setVar(label[1], d.getRadius(d.radiusSliderOut))
        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)

    @classmethod
    def getView(self):
        return "wiz_volume_mask_radii"


class ParticleMaskRadiusWizard(MaskRadiusWizard):
    pass


class VolumeMaskRadiusWizard(MaskRadiusWizard):
    pass


class ParticlesMaskRadiiWizard(MaskRadiiWizard):
    pass


class VolumeMaskRadiiWizard(MaskRadiiWizard):
    pass


class FilterWizard(EmWizard):

    def show(self, form, value, label, mode, unit=emcts.UNIT_PIXEL, **args):
        protocol = form.protocol
        provider = self._getProvider(protocol)

        if provider is not None:
            self.mode = mode
            args.update({'mode': self.mode,
                         'unit': unit})
            if self.mode == emcts.FILTER_LOW_PASS:
                args['showLowFreq'] = False
                args['highFreq'] = value[1]
                args['freqDecay'] = value[2]
            elif self.mode == emcts.FILTER_HIGH_PASS:
                args['showHighFreq'] = False
                args['lowFreq'] = value[0]
                args['freqDecay'] = value[2]
            elif self.mode == emcts.FILTER_BAND_PASS:
                args['lowFreq'] = value[0]
                args['highFreq'] = value[1]
                args['freqDecay'] = value[2]
            else:
                raise Exception("Unknown mode '%s'" % self.mode)

            d = BandPassFilterDialog(form.root, provider, **args)

            if d.resultYes():
                def setFormValue(flag, value, index):
                    if args.get(flag, True):
                        if unit == emcts.UNIT_ANGSTROM:
                            value = d.samplingRate / value
                        form.setVar(label[index], value)

                setFormValue('showLowFreq', d.getLowFreq(), 0)
                setFormValue('showHighFreq', d.getHighFreq(), 1)
                setFormValue('showDecay', d.getFreqDecay(), 2)

        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)


class FilterParticlesWizard(FilterWizard):
    pass


class FilterVolumesWizard(FilterWizard):
    pass


class GaussianWizard(EmWizard):

    def show(self, form, value, label, units=emcts.UNIT_PIXEL_FOURIER):
        protocol = form.protocol
        provider = self._getProvider(protocol)

        if provider is not None:
            args = {'freqSigma': value,
                    'unit': units
                    }

            d = GaussianFilterDialog(form.root, provider, **args)
            if d.resultYes():
                form.setVar(label, d.getFreqSigma())
        else:
            dialog.showWarning("Empty input", "Select elements first", form.root)


class GaussianParticlesWizard(GaussianWizard):
    pass


class GaussianVolumesWizard(GaussianWizard):
    pass


class PDBVolumeWizard(EmWizard):
    def show(self, form):
        if form.protocol.inputPDB.hasValue():
            pdb = form.protocol.inputPDB.get()
            print("pdb ", pdb._volume, str(pdb._volume))
            if pdb._volume:
                print("Setting ", str(pdb.getVolume()))
                ptr = pwobj.Pointer()
                ptr.copy(form.protocol.inputPDB)
                ptr.setExtended(ptr.getExtended() + "._volume")
                #                 ptr.set(pdb.getVolume())
                form.setVar('inputVol', ptr)


class ColorScaleWizardBase(EmWizard):
    """ Base wizard to edit color scale parameters
    Usage:
        1.- define new class inheriting this one
        2.- declare _targets = ColorScaleWizardBase.defineTargets(youviewer) in a class scope, right after class definition.
        3.- call ColorScaleWizardBase.defineColorScaleParams(group) in your viewer._defineParams
        4.- use attributes in your plotting method
    """

    @classmethod
    def defineTargets(cls, *viewersClass):
        """ :return targets list per each viewer class passed"""
        targets = []
        for viewer in viewersClass:
            targets.append((viewer, [HIGHEST_ATTR, LOWEST_ATTR, INTERVAL_ATTR, COLORMAP_ATTR]))

        return targets

    def show(self, form):
        viewer = form.protocol

        d = ColorScaleDialog(form.root, viewer.lowest.get(), viewer.highest.get(), viewer.intervals.get(),
                             viewer.colorMap.get())

        # If accepted
        if d.resultYes():
            form.setVar(LOWEST_ATTR, d.getLowest())
            form.setVar(HIGHEST_ATTR, d.getHighest())
            form.setVar(INTERVAL_ATTR, d.getIntervals())
            form.setVar(COLORMAP_ATTR, d.getColorPalette())

    @staticmethod
    def defineColorScaleParams(form, defaultHighest=10, defaultLowest=0, defaultIntervals=11, defaultColorMap="jet"):

        line = form.addLine("Color scale options:",
                            help="Options to define the color scale limits, intervals (advanced) and color set. Useful when you have outliers ruining your "
                                 "visualization/plot.")
        line.addParam(HIGHEST_ATTR, FloatParam, default=defaultHighest,
                      label="Highest",
                      help="Highest value for the scale")

        line.addParam(LOWEST_ATTR, FloatParam, default=defaultLowest,
                      label="Lowest",
                      help="lowest value of the scale.")

        line.addParam(INTERVAL_ATTR, IntParam, default=defaultIntervals,
                      label="Intervals",
                      help="Number of labels of the scale",
                      expertLevel=LEVEL_ADVANCED)

        line.addParam(COLORMAP_ATTR, StringParam, default=defaultColorMap,
                      label="Color set",
                      help="Combination of color for the scale")


# ===============================================================================
#  Dialogs used by wizards
# ===============================================================================

class PreviewDialog(dialog.Dialog):
    """ This will be the base class for several wizards.
    The layout of this wizard will be:
    1. Left panel(Items) that contains a list of items to preview
    2. Right-top panel (Preview) where some preview of the items will be displayed
    3. Right-bottom panel (Controls) where some controls can change the preview
    """

    def __init__(self, parent, provider, **args):
        """
        Params:
            parent: parent windows of the dialog.
            provider: the TreeProvider to populate items tree.
        """
        # Set the attributes in **args
        for k, v in args.items():
            setattr(self, k, v)

        self.provider = provider
        self.firstItem = provider.getObjects()[0]
        self.isInnerRad = False
        self.isMakingBigger = False
        self.step = 1
        self.expText = []
        buttons = [('Select', dialog.RESULT_YES),
                   ('Cancel', dialog.RESULT_CANCEL)]
        dialog.Dialog.__init__(self, parent, "Wizard",
                               buttons=buttons, default='Select', **args)

    def body(self, bodyFrame):
        bodyFrame.config()
        bodyFrame.columnconfigure(0, weight=1)
        bodyFrame.columnconfigure(1, weight=1)
        bodyFrame.columnconfigure(2, weight=1)

        # Create explanation label
        self.expText = ExplanationText(bodyFrame)
        self.expText.text.grid(row=0, column=0, columnspan=2, padx=5, pady=5, sticky='news')

        # Create items frame
        itemsFrame = tk.Frame(bodyFrame, bg=pw.TK_GRAY_DEFAULT)
        itemsFrame.grid(row=1, column=0, padx=5, sticky='news')
        itemsFrame.columnconfigure(0, weight=1)
        itemsFrame.rowconfigure(0, weight=1)
        itemsTree = BoundTree(itemsFrame, self.provider)
        itemsTree.grid(row=0, column=0, padx=5, pady=5, sticky='news')
        itemsTree.itemClick = self._itemSelected

        # Create preview frame
        previewFrame = tk.Frame(bodyFrame, bg=pw.TK_GRAY_DEFAULT)
        previewFrame.grid(row=1, column=1, padx=5, pady=5)
        self._beforePreview()
        self._createPreview(previewFrame)

        # Create controls frame
        controlsFrame = tk.Frame(bodyFrame)
        controlsFrame.grid(row=2, column=0, columnspan=2, padx=5, pady=5, sticky='sew')
        controlsFrame.columnconfigure(0, weight=1)
        controlsFrame.rowconfigure(2, weight=1)

        self._createControls(controlsFrame)
        self._itemSelected(self.firstItem)
        itemsTree.selectChildByIndex(0)  # Select the first item

    def _beforePreview(self):
        """ Called just before setting the preview.
        This is the place to set data values such as: labels, constants...
        """
        pass

    def _createPreview(self, frame):
        """ Should be implemented by subclasses to
        create the items preview.
        """
        pass

    def _createControls(self, frame):
        """ Create controls to be used. """
        pass

    def _itemSelected(self, obj):
        """ This will be call when an item in the tree is selected. """
        pass


class ImagePreviewDialog(PreviewDialog):

    def _beforePreview(self):
        self.dim = 256
        self.previewLabel = ''

    def _createPreview(self, frame):
        """ Should be implemented by subclasses to
        create the items preview.
        """
        from pyworkflow.gui.matplotlib_image import ImagePreview
        self.preview = ImagePreview(frame, self.dim, label=self.previewLabel)
        self.preview.grid(row=0, column=0)

    def _itemSelected(self, obj):

        index = obj.getIndex()
        filename = emlib.image.ImageHandler.fixXmippVolumeFileName(obj)
        if index:
            filename = "%03d@%s" % (index, filename)

        self.image = emlib.image.ImageHandler()._img

        try:
            self.image.readPreview(filename, self.dim)
            if filename.endswith('.psd'):
                self.image.convertPSD()
            self.Z = self.image.getData()
        except Exception as e:
            from pyworkflow.gui.matplotlib_image import getPngData
            self.Z = getPngData(pw.findResource('no-image.gif'))
            dialog.showError("Input particles", "Error reading image <%s>"
                             % filename, self)
        self.preview.updateData(self.Z)


class DownsampleDialog(ImagePreviewDialog):

    def _beforePreview(self):
        self.expText.updateExpText(emcts.FREQ_BANDPASS_WIZ_MSG, width=75)
        ImagePreviewDialog._beforePreview(self)
        self.lastObj = None
        self.rightPreviewLabel = "PSD"
        self.message = "Computing PSD..."
        self.previewLabel = "Micrograph"
        self.rightImage = emlib.image.ImageHandler()._img

    def _createPreview(self, frame):
        """ Should be implemented by subclasses to
        create the items preview.
        """
        leftFrame = tk.Frame(frame)
        leftFrame.grid(row=0, column=0)

        rightFrame = tk.Frame(frame)
        rightFrame.grid(row=0, column=1)

        ImagePreviewDialog._createPreview(self, leftFrame)
        self.rightPreview = self._createRightPreview(rightFrame)
        self.rightPreview.grid(row=0, column=0)

    def _createRightPreview(self, rightFrame):
        from pyworkflow.gui.matplotlib_image import ImagePreview
        return ImagePreview(rightFrame, self.dim, label=self.rightPreviewLabel)

    def _createControls(self, frame):
        self.downVar = tk.StringVar()
        self.downVar.set(getattr(self, 'downsample', 1))
        downFrame = tk.Frame(frame)
        downFrame.grid(row=0, column=0, pady=5, sticky='new')
        downLabel = tk.Label(downFrame, text='Downsample')
        downLabel.grid(row=0, column=0, sticky='ew')
        downEntry = tk.Entry(downFrame, width=10, textvariable=self.downVar)
        downEntry.grid(row=0, column=1, padx=5, sticky='ew')
        downButton = tk.Button(downFrame, text='Preview', command=self._doPreview)
        downButton.grid(row=1, column=1, padx=5, pady=5, sticky='ew')

    def getDownsample(self):
        return float(self.downVar.get())

    def manageMaskVals(self):
        # To be coded by each child, if necessary
        pass

    def _itemSelected(self, obj):
        self.lastObj = obj
        ImagePreviewDialog._itemSelected(self, obj)

        dialog.FlashMessage(self, self.message, func=self._computeRightPreview)
        self.rightPreview.updateData(self.rightImage.getData())
        self.manageMaskVals()

    def _doPreview(self, e=None):
        if self.lastObj is None:
            dialog.showError("Empty selection",
                             "Select an item first before preview", self)
        else:
            self._itemSelected(self.lastObj)

    def _computeRightPreview(self):
        """ This function should compute the right preview
        using the self.lastObj that was selected
        """
        emlib.fastEstimateEnhancedPSD(self.rightImage,
                                      self.lastObj.getFileName(),
                                      self.getDownsample(), self.dim, 2)


class CtfDialog(DownsampleDialog):

    def _createRightPreview(self, rightFrame):
        self.isInnerRad = True
        listeners = {"<Button-4>": self.makeBigger, "<Button-5>": self.makeSmaller,
                     "<Up>": self.upKeyPress, "<Down>": self.downKeyPress}
        from pyworkflow.gui.matplotlib_image import PsdPreview
        return PsdPreview(rightFrame, self.dim, self.lf, self.hf,
                          label=self.rightPreviewLabel, listenersDict=listeners)

    def downKeyPress(self, event):
        self.isInnerRad = False
        self.highlightOuterSlider()

    def upKeyPress(self, event):
        self.isInnerRad = True
        self.highlightInnerSlider()

    def highlightOuterSlider(self):
        self.hfSlider.highlightLabel()
        self.lfSlider.removeHighlightFromLabel()

    def highlightInnerSlider(self):
        self.lfSlider.highlightLabel()
        self.hfSlider.removeHighlightFromLabel()

    def makeBigger(self, event):
        self.isMakingBigger = True
        if self.isInnerRad:
            self.lf = self.lf + self.step
        else:
            new_val = self.hf + self.step
            if new_val <= 0.5:  # Don't make the mask bigger unless the is equal or lower than the max
                self.hf = new_val
        self.manageMaskVals()

    def makeSmaller(self, event):
        self.isMakingBigger = False
        if self.isInnerRad:
            new_val = self.lf - self.step
            if new_val >= 0:
                self.lf = new_val
        else:
            self.hf = self.hf - self.step
        self.manageMaskVals()

    def manageMaskVals(self):
        if self.isMakingBigger:
            if self.isInnerRad and self.lf >= self.hf:  # Inner ring can't be bigger than outer ring
                # Subtract one step to go back to the nearest lower value
                self.lf = self.hf - self.step
        else:
            if not self.isInnerRad and self.hf <= self.lf:  # Outer ring can't be smaller than inner
                # ring
                # Add one step to go back to the nearest higher value
                self.hf = self.lf + self.step

        # Set the ring sliders in case it comes from the mouse wheel
        self.setFreq(self.lfSlider, self.lf)
        self.setFreq(self.hfSlider, self.hf)

        # Show values
        self.showValues(self.hfVar, self.hfSlider)
        self.showValues(self.lfVar, self.lfSlider)

        # Update mask
        self.rightPreview.updateFreq(self.lf, self.hf)

    def _showInAngstroms(self):
        return getattr(self, 'showInAngstroms', False)

    def _createControls(self, frame):
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)
        frame.columnconfigure(2, weight=1)

        self.step = 0.01
        self.freqFrame = ttk.LabelFrame(frame, text="Frequencies")
        self.freqFrame.columnconfigure(0, weight=3)
        self.freqFrame.grid(row=0, column=2, sticky='news')
        self.lfSlider = LabelSlider(self.freqFrame, 'Low freq',
                                    value=self.lf,
                                    from_=0.01,
                                    to=0.5,
                                    length=178,
                                    labelWidth=13,
                                    step=self.step,
                                    showvalue=0,
                                    callback=lambda a, b, c: self.updateSliderInnerRadius(),
                                    )
        self.lfSlider.grid(row=0, column=0, padx=5, pady=5, sticky='news')
        self.lfSlider.highlightLabel()
        self.hfSlider = LabelSlider(self.freqFrame, 'High freq',
                                    value=self.lf,
                                    from_=0.01,
                                    to=0.5,
                                    length=178,
                                    labelWidth=13,
                                    step=self.step,
                                    showvalue=0,
                                    callback=lambda a, b, c: self.updateSliderOuterRadius(),
                                    )
        self.hfSlider.grid(row=1, column=0, padx=5, pady=5, sticky='news')

        # Pack and configure low freq slider
        self.lfVar = tk.StringVar()
        self.lfLabel = tk.Label(self.freqFrame, textvariable=self.lfVar, width=26)
        self.lfLabel.grid(row=0, column=1, sticky='nse', padx=5, pady=5)
        # Pack and configure high freq slider
        self.hfVar = tk.StringVar()
        self.hfLabel = tk.Label(self.freqFrame, textvariable=self.hfVar, width=26)
        self.hfLabel.grid(row=1, column=1, sticky='nse', padx=5, pady=5)
        # Update both mask and sliders with the initial values
        self.manageMaskVals()

    def getDownsample(self):
        return 1.0  # Micrograph previously downsample, not taken into account here

    def updateFreqRing(self):
        self.rightPreview.updateFreq(self.getLowFreq(), self.getHighFreq())
        self.showValues(self.lfVar, self.lfSlider)
        self.showValues(self.hfVar, self.hfSlider)

    def updateSliderOuterRadius(self):
        new_val = self.getFreq(self.hfSlider)
        # Carry out the action only if the slider is clicked & dragged (prevent from minimal variations when the mouse
        # is on the slider but not clicking on it)
        if abs(new_val - self.hf) >= self.step:
            self.highlightOuterSlider()
            self.isInnerRad = False
            self.isMakingBigger = False
            # Check if the user is making the ring bigger
            if new_val > self.hf:
                self.isMakingBigger = True

            self.hf = new_val
            self.manageMaskVals()

    def updateSliderInnerRadius(self):
        new_val = self.getFreq(self.lfSlider)
        # Highlight only if the slider is clicked & dragged (prevent from minimal variations when the mouse is on the
        # slider but not clicking on it)
        if abs(new_val - self.lf) >= self.step:
            self.highlightInnerSlider()
            self.isInnerRad = True
            self.isMakingBigger = False
            # Check if the user is making the ring bigger
            if new_val > self.lf:
                self.isMakingBigger = True

            self.lf = new_val
            self.manageMaskVals()

    def showValues(self, var2set, labSlider):
        """
        Show the values selected for the inner and outer radius. If the units are angstroms (sampling_rate = 1,
        it will show only one value to avoid redundancies
        """
        sr = self.firstItem.getSamplingRate()
        freqVal = max(self.getFreq(labSlider), 0.0001)  # avoid division by zero
        if sr == 1:
            var2set.set('{:2.2f} {}'.format(labSlider.slider.get(), emcts.UNIT_PIXEL))
        else:
            var2set.set('{:2.2f} rad/Å | {:5.1f} Å'.format(labSlider.slider.get(),
                                                           self.getDownsample() * sr / freqVal))

    def getLowFreq(self):
        return self.lfSlider.get()

    def getHighFreq(self):
        return self.hfSlider.get()

    @staticmethod
    def getFreq(freqSlider):
        return freqSlider.get()

    @staticmethod
    def setFreq(freqSlider, val):
        freqSlider.slider.set(val)


class CtfDownsampleDialog(CtfDialog):

    def _createControls(self, frame):
        DownsampleDialog._createControls(self, frame)
        CtfDialog._createControls(self, frame)

    def getDownsample(self):
        return float(self.downVar.get())


class BandPassFilterDialog(DownsampleDialog):

    def _beforePreview(self):
        ImagePreviewDialog._beforePreview(self)
        self.lastObj = None
        self.rightPreviewLabel = "Filtered"
        self.message = "Computing filtered image..."
        self.previewLabel = "Image"
        self.rightImage = emlib.image.ImageHandler()._img

    def _createControls(self, frame):
        self.freqFrame = ttk.LabelFrame(frame,
                                        text="Frequencies (%s)" % self.unit,
                                        padding="5 5 5 5")
        self.freqFrame.grid(row=0, column=0)

        self.showLowFreq = getattr(self, 'showLowFreq', True)
        self.showHighFreq = getattr(self, 'showHighFreq', True)
        self.showDecay = getattr(self, 'showDecay', True)

        if (not self.showLowFreq) or (not self.showHighFreq):
            label_high = 'Freq'
            label_low = 'Freq'
        else:
            label_high = 'High freq'
            label_low = 'Low freq'

        self.samplingRate = 1.0
        self.sliTo = 0.5
        self.sliFrom = 0.01
        if self.unit == emcts.UNIT_ANGSTROM:
            self.samplingRate = self.firstItem.getSamplingRate()
            self.itemDim, _, _ = self.firstItem.getDim()
            self.sliFrom = 2. * self.samplingRate
            self.sliTo = 2. * self.itemDim * self.samplingRate

        self.step = self.sliTo / 1000

        if self.showLowFreq:
            self.lfSlider = self.addFreqSlider(label_low, self.lowFreq, col=0)
        if self.showHighFreq:
            self.hfSlider = self.addFreqSlider(label_high, self.highFreq, col=1)
        if self.showDecay:
            self.freqDecaySlider = self.addFreqSlider('Decay',
                                                      self.freqDecay, col=2)

    def addFreqSlider(self, label, value, col):
        fromValue = self.sliFrom
        toValue = self.sliTo
        if self.unit == emcts.UNIT_ANGSTROM:
            fromValue = self.sliTo
            toValue = self.sliFrom
        slider = LabelSlider(self.freqFrame, label, from_=fromValue, to=toValue,
                             step=self.step, value=value,
                             callback=lambda a, b, c: self.updateFilteredImage())
        slider.grid(row=0, column=col, padx=5, pady=5)
        return slider

    def updateFilteredImage(self):
        self._computeRightPreview()
        self.rightPreview.updateData(self.rightImage.getData())

    def _computeRightPreview(self):
        """ This function should compute the right preview
        using the self.lastObj that was selected
        """
        emlib.bandPassFilter(self.rightImage,
                             emlib.image.ImageHandler.locationToXmipp(self.lastObj),
                             self.getLowFreq(), self.getHighFreq(),
                             self.getFreqDecay(), self.dim)

    def getLowFreq(self):
        if self.showLowFreq:
            if self.unit == emcts.UNIT_ANGSTROM:
                return self.samplingRate / self.lfSlider.get()
            else:
                return self.lfSlider.get()
        return 0.01

    def getHighFreq(self):
        if self.showHighFreq:
            if self.unit == emcts.UNIT_ANGSTROM:
                return self.samplingRate / self.hfSlider.get()
            else:
                return self.hfSlider.get()
        return 1.0

    def getFreqDecay(self):
        if self.showDecay:
            if self.unit == emcts.UNIT_ANGSTROM:
                return self.samplingRate / self.freqDecaySlider.get()
            else:
                return self.freqDecaySlider.get()
        return 0.01


class GaussianFilterDialog(BandPassFilterDialog):

    def _createControls(self, frame):
        self.freqVar = tk.StringVar()
        self.freqVar.set(getattr(self, 'freqSigma', 1))
        freqFrame = tk.Frame(frame)
        freqFrame.grid(row=0, column=0, sticky='nw')
        downEntry = tk.Entry(freqFrame, width=10, textvariable=self.freqVar)
        downEntry.grid(row=0, column=1, padx=5, pady=5)
        downButton = tk.Button(freqFrame, text='Preview',
                               command=self._doPreview)
        downButton.grid(row=0, column=2, padx=5, pady=5)

    def getFreqSigma(self):
        return float(self.freqVar.get())

    def _computeRightPreview(self):
        """ This function should compute the right preview
        using the self.lastObj that was selected
        """
        emlib.gaussianFilter(self.rightImage,
                             emlib.image.ImageHandler.locationToXmipp(self.lastObj),
                             self.getFreqSigma(), self.dim)


class MaskPreviewDialog(ImagePreviewDialog):

    def _beforePreview(self):
        self.dim = 256
        self.unit = getattr(self, 'unit', emcts.UNIT_PIXEL)
        self.samplingRate = self.firstItem.getSamplingRate()
        self.dim_par = self.firstItem.getDim()[0]
        self.ratio = self.dim / float(self.dim_par)
        self.previewLabel = 'Central slice'

    def _createPreview(self, frame):
        """ Should be implemented by subclasses to
        create the items preview.
        """

        # Insert the corresponding explanation text
        self.expText.updateExpText(emcts.CIRCLE_MASK_WIZ_MSG)

        from pyworkflow.gui.matplotlib_image import MaskPreview

        if self.maskRadius == -1:
            self.iniRadius = self.dim_par / 2
        else:
            self.iniRadius = self.maskRadius

        if self.unit == emcts.UNIT_ANGSTROM:
            self.iniRadius = self.iniRadius / self.samplingRate

        listeners = {"<Button-4>": self.makeBigger, "<Button-5>": self.makeSmaller}

        self.preview = MaskPreview(frame, self.dim, label=self.previewLabel,
                                   listenersDict=listeners)

        self.preview.grid(row=1, column=0)

    def makeBigger(self, event):
        new_val = self.iniRadius + 1
        if new_val <= self.dim_par / 2:  # Don't make the mask bigger unless the is equal or lower than the max
            self.iniRadius = self.iniRadius + 1
            self.manageMaskVals()

    def makeSmaller(self, event):
        new_val = self.iniRadius - 1
        if new_val >= 1:  # Don't make the mask smaller unless the radius is equal or greater than 1
            self.iniRadius = self.iniRadius - 1
            self.manageMaskVals()

    def manageMaskVals(self):
        # Set the ring slider in case it comes from the mouse wheel
        self.setRadius(self.radiusSlider, self.iniRadius)
        # Show values
        self.showValues(self.hfVar, self.radiusSlider)
        # Update mask
        self.preview.updateMask(self.iniRadius * self.ratio)

    def _createControls(self, frame):
        self.addRadiusBox(frame)
        self.hfVar = tk.StringVar()
        self.hfLabel = tk.Label(frame, textvariable=self.hfVar, width=20)
        self.hfLabel.grid(row=0, column=1, sticky='NSE', padx=5, pady=5)
        self.manageMaskVals()

    def addRadiusBox(self, parent):
        self.radiusSlider = LabelSlider(parent, 'Mask radius',
                                        from_=1, to=int(self.dim_par / 2),
                                        value=self.iniRadius,
                                        step=1,
                                        length=150,
                                        showvalue=0,
                                        callback=lambda a, b, c: self.updateSliderRadius())
        self.radiusSlider.grid(row=0, column=0, sticky='NEWS')

    def updateSliderRadius(self):

        new_val = self.getRadius(self.radiusSlider)
        # Carry out the action only if the slider is clicked & dragged (prevent from minimal variations when the mouse
        # is on the slider but not clicking on it)
        if abs(new_val - self.iniRadius) >= 1:
            self.iniRadius = new_val
            self.manageMaskVals()

    def showValues(self, var2set, radiusSlider):
        """
        Show the values selected for the inner and outer radius. If the units are angstroms (sampling_rate = 1,
        it will show only one value to avoid redundancies
        """
        pixVal = self.getRadius(radiusSlider)
        if self.samplingRate == 1:
            var2set.set('{:6.1f} {}'.format(pixVal, emcts.UNIT_PIXEL))
        else:
            var2set.set('{:5.0f} pix | {:6.1f} Å'.format(pixVal,
                                                          self.getRadiusAngstroms(radiusSlider),
                                                          ))

    @staticmethod
    def getRadius(radiusSlider):
        return radiusSlider.get()

    @staticmethod
    def setRadius(radiusSlider, val):
        radiusSlider.slider.set(val)

    def getRadiusAngstroms(self, radiusSlider):
        return convertPixToLength(self.samplingRate, radiusSlider.get())


class MaskRadiiPreviewDialog(MaskPreviewDialog):

    def _createPreview(self, frame):
        """ Should be implemented by subclasses to
        create the items preview.
        """
        # Insert the corresponding explanation text
        self.expText.updateExpText(emcts.RING_MASK_WIZ_MSG)

        from pyworkflow.gui.matplotlib_image import MaskPreview
        if self.innerRadius is None:
            self.innerRadius = 1
        if self.outerRadius is None or self.outerRadius == -1 or self.outerRadius > self.dim_par / 2:
            self.outerRadius = int(self.dim_par / 2)

        if self.unit == emcts.UNIT_ANGSTROM:
            self.innerRadius = self.innerRadius / self.samplingRate
            self.outerRadius = self.innerRadius / self.samplingRate

        listeners = {"<Button-4>": self.makeBigger, "<Button-5>": self.makeSmaller,
                     "<Up>": self.upKeyPress, "<Down>": self.downKeyPress}

        self.preview = MaskPreview(frame, self.dim,
                                   label=self.previewLabel,
                                   listenersDict=listeners)

        self.preview.grid(row=1, column=0)

    def upKeyPress(self, event):
        self.isInnerRad = False
        self.highlightOuterSlider()

    def downKeyPress(self, event):
        self.isInnerRad = True
        self.highlightInnerSlider()

    def highlightOuterSlider(self):
        self.radiusSliderOut.highlightLabel()
        self.radiusSliderIn.removeHighlightFromLabel()

    def highlightInnerSlider(self):
        self.radiusSliderIn.highlightLabel()
        self.radiusSliderOut.removeHighlightFromLabel()

    def makeBigger(self, event):
        self.isMakingBigger = True
        if self.isInnerRad:
            self.innerRadius = self.innerRadius + self.step
        else:
            new_val = self.outerRadius + self.step
            if new_val <= int(self.dim_par / 2):  # Don't make the mask bigger unless the is equal or lower than the max
                self.outerRadius = new_val
        self.manageMaskVals()

    def makeSmaller(self, event):
        self.isMakingBigger = False
        if self.isInnerRad:
            new_val = self.innerRadius - self.step
            if new_val >= 0:
                self.innerRadius = new_val
        else:
            self.outerRadius = self.outerRadius - self.step
        self.manageMaskVals()

    def manageMaskVals(self):
        if self.isMakingBigger:
            if self.isInnerRad and self.innerRadius >= self.outerRadius:  # Inner ring can't be bigger than outer ring
                # Subtract one step to go back to the nearest lower value
                self.innerRadius = self.outerRadius - self.step
        else:
            if not self.isInnerRad and self.outerRadius <= self.innerRadius:  # Outer ring can't be smaller than inner
                # ring
                # Add one step to go back to the nearest higher value
                self.outerRadius = self.innerRadius + self.step

        # Set the ring sliders in case it comes from the mouse wheel
        self.setRadius(self.radiusSliderIn, self.innerRadius)
        self.setRadius(self.radiusSliderOut, self.outerRadius)

        # Show values
        self.showValues(self.orVar, self.radiusSliderOut)
        self.showValues(self.irVar, self.radiusSliderIn)

        # Update mask
        self.preview.updateMask(self.outerRadius * self.ratio,
                                self.innerRadius * self.ratio)

    def _createControls(self, frame):
        self.step = 1
        to = int(self.dim_par / 2)
        self.radiusSliderOut = LabelSlider(frame, 'Outer radius',
                                           from_=1, to=to,
                                           value=self.outerRadius, step=self.step,
                                           callback=lambda a, b, c: self.updateSliderOuterRadius(),
                                           length=150, showvalue=0)
        self.radiusSliderOut.highlightLabel()
        self.orVar = tk.StringVar()
        self.orLabel = tk.Label(frame, textvariable=self.orVar, width=20)

        self.radiusSliderIn = LabelSlider(frame, 'Inner radius',
                                          from_=1, to=to,
                                          value=self.innerRadius, step=self.step,
                                          callback=lambda a, b, c: self.updateSliderInnerRadius(),
                                          length=150, showvalue=0)
        self.irVar = tk.StringVar()
        self.irLabel = tk.Label(frame, textvariable=self.irVar, width=20)
        # Pack and configure outer radius slider
        self.radiusSliderOut.grid(row=0, column=0, sticky='NEWS')
        self.radiusSliderOut.columnconfigure(0, weight=1)
        self.orLabel.grid(row=0, column=1, sticky='NSE', padx=5, pady=5)
        self.orLabel.columnconfigure(0, weight=1)
        # Pack and configure inner radius slider
        self.radiusSliderIn.grid(row=1, column=0, sticky='NEWS')
        self.radiusSliderIn.columnconfigure(0, weight=1)
        self.irLabel.grid(row=1, column=1, sticky='NSE', padx=5, pady=5)
        self.irLabel.columnconfigure(0, weight=1)
        # Update both mask and sliders with the initial values
        self.manageMaskVals()

    def updateSliderOuterRadius(self):
        new_val = self.getRadius(self.radiusSliderOut)
        # Carry out the action only if the slider is clicked & dragged (prevent from minimal variations when the mouse
        # is on the slider but not clicking on it)
        if abs(new_val - self.outerRadius) >= self.step:
            self.highlightOuterSlider()
            self.isInnerRad = False
            self.isMakingBigger = False
            # Check if the user is making the ring bigger
            if new_val > self.outerRadius:
                self.isMakingBigger = True

            self.outerRadius = new_val
            self.manageMaskVals()

    def updateSliderInnerRadius(self):
        new_val = self.getRadius(self.radiusSliderIn)
        # Highlight only if the slider is clicked & dragged (prevent from minimal variations when the mouse is on the
        # slider but not clicking on it)
        if abs(new_val - self.innerRadius) >= self.step:
            self.highlightInnerSlider()
            self.isInnerRad = True
            self.isMakingBigger = False
            # Check if the user is making the ring bigger
            if new_val > self.innerRadius:
                self.isMakingBigger = True

            self.innerRadius = new_val
            self.manageMaskVals()


class ColorScaleDialog(dialog.Dialog):
    """ This will assist users to choose the color scale and range for
    local resolution viewers
    """

    def __init__(self, parentWindow, lowest, highest, intervals, colorPalette):
        """
            :param highest: highest resolution value for the scale
            :param lowest: lowest resolution value for the scale
            :param intervals: number of labels for the scale
            :param colorPalette: color palette to use
        """
        self.highest = tk.DoubleVar(value=highest)
        self.lowest = tk.DoubleVar(value=lowest)
        self.intervals = tk.IntVar(value=intervals)
        self.colorPalette = tk.StringVar(value=colorPalette)
        self.info = tk.StringVar()

        # GUI attributes initialization
        self.palette = None
        self.params = None

        dialog.Dialog.__init__(self, parentWindow, "Color scale wizard", default="None")

    # Getters
    def getIntervals(self):
        return int(self.intervals.get())

    def getHighest(self):
        return float(self.highest.get())

    def getLowest(self):
        return float(self.lowest.get())

    def getColorPalette(self):
        return self.colorPalette.get()

    ### ----- GUI methods ------ ###
    def body(self, master):
        """ Draws the main frame of the dialog"""
        body = tk.Frame(self)
        body.grid(row=0, column=0, sticky="news")
        body.grid_columnconfigure(1, weight=10)

        # GUI attributes
        self.palette = tk.Frame(body)
        self.palette.grid(row=0, column=0, sticky='nes',
                          padx=5, pady=5)

        # Params
        self.params = tk.Frame(body)
        self.params.grid(row=0, column=1, sticky='news',
                         padx=5, pady=5)
        self.params.bind("<Key>", self._keyPressedOnParams)

        self._drawPalette()
        self._fillParams()

    def _drawLabel(self, color, value, count):
        label = tk.Label(self.palette, text=value)
        label.config(bg=color)
        label.grid(row=count, column=0, sticky='news')

    def _fillParams(self):

        # Add highest scale value
        highValueLabel = tk.Label(self.params, text="Highest:")
        highValueLabel.grid(row=0, column=0, sticky="e")
        self._addEntry(0, 1, self.highest)

        # Add lowest value
        lowValueLabel = tk.Label(self.params, text="Lowest:")
        lowValueLabel.grid(row=1, column=0, sticky="e")
        self._addEntry(1, 1, self.lowest)

        intervalsLabel = tk.Label(self.params, text="Intervals:")
        intervalsLabel.grid(row=2, column=0, sticky="e")
        self._addEntry(2, 1, self.intervals)

        # Palettes
        paletteLabel = tk.Label(self.params, text="Color set:")
        paletteLabel.grid(row=3, column=0, sticky="e")

        availablePalettes = self.getAvailablePalettes()
        opt = ttk.Combobox(self.params, textvariable=self.colorPalette, values=availablePalettes)
        opt.grid(row=3, column=1, sticky="news")
        self.colorPalette.trace("w", self._paletteChanged)

        # Label to store information: number of color of the color map, e.g.
        infoLabel = tk.Label(self.params, textvariable=self.info, text="")
        infoLabel.grid(row=4, column=0, columnspan=2, sticky="news")

    def _addEntry(self, row, column, var):
        newEntry = tk.Entry(self.params, textvariable=var)
        newEntry.grid(row=row, column=column, sticky="news")
        newEntry.lift()
        # Bind events
        newEntry.bind("<Return>", self._paramChanged)
        newEntry.bind("<FocusIn>", self._selectAllText)
        newEntry.bind("<Button-1>", self._selectAllText)
        newEntry.bind("<FocusOut>", self._paramChanged)

    def _paletteChanged(self, *args):
        self._drawPalette()
        info = "%s has %s colors." % (self.getColorPalette(), cm.get_cmap(self.getColorPalette()).N)
        self.info.set(info)

    def _drawPalette(self):
        """ Draws the palette using current values: palette, highest, lowest and intervals"""
        # clean first
        for widget in self.palette.winfo_children():
            widget.destroy()

        # get the colors
        colorList = getHexColorList(self.getIntervals(), colorName=self.getColorPalette())

        # Get the label values
        labelValues = splitRange(self.getLowest(), self.getHighest(), self.getIntervals())

        # Draw it
        count = 0
        for color, value in zip(colorList, labelValues):
            self._drawLabel(color, value, count)
            count = count + 1

    ### Events handling
    def _paramChanged(self, event):
        # Handles change in palette Option
        self._drawPalette()

    def _keyPressedOnParams(self, event):
        # handles keypress on params frame
        print("PASA key press")
        self.colorPalette.set("jet")

    def getAvailablePalettes(self):
        """ Returns a list of all available palettes"""
        return list(cm.cmap_d.keys())

    def _selectAllText(self, event):
        """ Select all the text of the widget that triggered the event"""
        # select text
        event.widget.select_range(0, 'end')
        # move cursor to the end
        event.widget.icursor('end')
        # stop propagation
        # return 'break'


class FormulaDialog(dialog.Dialog):
    """ This will assist users to create a formula based on class attibutes.
    """

    def __init__(self, parentWindow, set, formula=""):
        """
            :param set: set with the items to use in the formulae
            :param formula: initial formulate to load
        """
        self.set = set
        self.item = set.getFirstItem()
        self.formula = tk.StringVar(value=formula)
        self.formula.trace("w", self.evaluateFormula)
        self.formulaTxt = None # To be instantiated later.
        self.info = tk.StringVar()
        dialog.Dialog.__init__(self, parentWindow, "Formula wizard", default="None")


    # Getters
    def getFormula(self):
        return self.formula.get()

    def getItem(self):
        return self.item

    ### ----- GUI methods ------ ###
    def _createTree(self, parent):
        provider = AttributesTreeProvider(self.item)
        self.tree = BoundTree(parent, provider)
        self.tree.itemDoubleClick = self.addAttributeToFormula
        self.tree.grid(row=1, column=0)

    def addAttributeToFormula(self, event):
        attr = self.tree.getSelectedObjects()[0]
        self._insertText("item." + attr.attrName , self.formulaTxt)

    def body(self, master):
        """ Draws the main frame of the dialog"""
        body = tk.Frame(self)
        body.grid(row=0, column=0, sticky="news")
        body.grid_columnconfigure(1, weight=10)

        # GUI attributes
        self.attributes = tk.Frame(body)
        self.attributes.grid(row=0, column=0, sticky='nes',
                          padx=5, pady=5)

        # Formula
        self.formulaFrame = tk.Frame(body)
        self.formulaFrame.grid(row=0, column=1, sticky='news',
                         padx=5, pady=5)
        #self.params.bind("<Key>", self._keyPressedOnParams)

        self._fillParams()


    def _fillParams(self):

        # Add the formula text
        formulaLbl = tk.Label(self.formulaFrame, text="Formula:")
        formulaLbl.grid(row=0, column=0, sticky="w")
        self.formulaTxt = self._addEntry(1, 0, self.formula,  width=100)

        self._createTree(self.attributes)

        # Label to show formula result or error
        infoLabel = tk.Label(self.formulaFrame, textvariable=self.info)
        infoLabel.grid(row=2, column=0, columnspan=2, sticky="news")

    def _addEntry(self, row, column, var, width=None):

        newEntry = tk.Entry(self.formulaFrame, textvariable=var, width=width)
        newEntry.grid(row=row, column=column, sticky="news")
        newEntry.lift()
        # Bind events
        newEntry.bind("<Return>", self._paramChanged)
        newEntry.bind("<FocusOut>", self._paramChanged)

        return newEntry

    def evaluateFormula(self, name='', index='', mode=''):
        try:
            item = self.item
            result = eval(self.formula.get())
            self.info.set(result)
        except Exception as e:
            self.info.set(str(e))

    ### Events handling
    def _paramChanged(self, event):
        # Handles change in palette Option
        self.evaluateFormula()

    def _selectAllText(self, event):
        """ Select all the text of the widget that triggered the event"""
        # select text
        event.widget.select_range(0, 'end')
        # move cursor to the end
        event.widget.icursor('end')
        # stop propagation
        # return 'break'

    def _insertText(self, text, widget):
        """ Insert the passed text in the entry at current cursor location"""
        # Get inserting position
        pos = widget.index(tk.INSERT)

        # Insert he text
        widget.insert(pos, text)

def insertText (target, textToInsert, position):
    """ Inserts a text into another at a position

    :param target: text to do the insertion on
    :param textToInsert: text to be inserted
    :param position: position where to insert the new text"""

    return target[:position] + textToInsert + target[position:]


import logging
import os
from subprocess import Popen

logger = logging.getLogger(__name__)
from datetime import datetime

import sqlite3
import tempfile
import time

import pyworkflow as pw
import metadataviewer
from metadataviewer.dao.model import IDAO
from metadataviewer.model import Table, Column, BoolRenderer, ImageRenderer, StrRenderer
from metadataviewer.model.renderers import ImageReader, Action

from functools import lru_cache


from PIL import Image
import mrcfile
import numpy as np


SCIPION_OBJECT_ID = "SCIPION_OBJECT_ID"
SCIPION_PORT = "SCIPION_PORT"


class MRCImageReader(ImageReader):
    @classmethod
    def open(cls, path:str):
        isVol = path.endswith(":mrc")

        path = path.replace(":mrc", "")
        if not "@" in path:
            path ="1@"+path
        filePath = path.split('@')

        index = int(filePath[0])
        fileName = filePath[-1]
        mrcImg = cls.getMrcImage(fileName)
        if mrcImg.is_volume() or isVol:
            dim = mrcImg.data.shape
            x = int(dim[0] / 2)
            imfloat = mrcImg.data[x, :, :]
        elif mrcImg.is_image_stack():
            imfloat = mrcImg.data[index-1]
        else:
            imfloat = mrcImg.data

        iMax = imfloat.max()
        iMin = imfloat.min()
        im255 = ((imfloat - iMin) / (iMax - iMin) * 255).astype(np.uint8)
        img = Image.fromarray(im255)
        return img


    @classmethod
    @lru_cache
    def getMrcImage(cls, fileName):
        logger.info("Reading %s" % fileName)
        return mrcfile.mmap(fileName, mode='r+')

    @classmethod
    def getCompatibleFileTypes(cls) -> list:
        return ['mrc', 'mrc:mrc', 'mrcs', 'em', 'rec', 'ali', 'st', 'mrcs:mrc',
                'mrcs:mrcs', 'mrc:mrcs']


class STKImageReader(ImageReader):
    IMG_BYTES = None
    stk_handler = None
    header_info = None
    HEADER_OFFSET = 1024
    FLOAT32_BYTES = 4
    TYPE = None

    @classmethod
    def open(cls, path):
        stk = path.split('@')
        if len(stk) > 1:
            image = cls.read(stk[-1], int(stk[0]))
            return image

    @classmethod
    def read(cls, filename, id):
        """
        Reads a given image
           :param filename (str) --> Image to be read
        """
        cls.stk_handler = open(filename, "rb")
        cls.header_info = cls.readHeader()
        cls.IMG_BYTES = cls.FLOAT32_BYTES * cls.header_info["n_columns"] ** 2
        image = cls.readImage(id - 1)
        iMax = image.max()
        iMin = image.min()
        image = ((image - iMin) / (iMax - iMin) * 255).astype('uint8')
        image = Image.fromarray(image)
        return image

    @classmethod
    def readHeader(cls):
        """
        Reads the header of the current file as a dictionary
            :returns The current header as a dictionary
        """
        header = cls.readNumpy(0, cls.HEADER_OFFSET)

        header = dict(img_size=int(header[1]), n_images=int(header[25]),
                      offset=int(header[21]),
                      n_rows=int(header[1]), n_columns=int(header[11]),
                      n_slices=int(header[0]),
                      sr=float(header[20]))

        cls.TYPE = "stack" if header["n_images"] > 1 else "volume"

        return header

    @classmethod
    def readNumpy(cls, start, end):
        """
        Read bytes between start and end as a Numpy array
            :param start (int) --> Start byte
            :param end (int) --> End byte
            :returns decoded bytes as Numpy array
        """
        return np.frombuffer(cls.readBinary(start, end), dtype=np.float32)

    @classmethod
    def readBinary(cls, start, end):
        """
        Read bytes between start and end
            :param start (int) --> Start byte
            :param end (int) --> End byte
            :returns the bytes read
        """
        cls.seek(start)
        return cls.stk_handler.read(end)

    @classmethod
    def readImage(cls, iid):
        """
        Reads a given image in the stack according to its ID
            :param iid (int) --> Image id to be read
            :returns Image as Numpy array
        """

        if cls.TYPE == "stack":
            start = 2 * cls.header_info["offset"] + iid * (
                    cls.IMG_BYTES + cls.header_info["offset"])
        else:
            start = cls.header_info["offset"] + iid * cls.IMG_BYTES

        img_size = cls.header_info["n_columns"]
        return cls.readNumpy(start, cls.IMG_BYTES).reshape([img_size, img_size])

    @classmethod
    def seek(cls, pos):
        """
        Move file pointer to a given position
            :param pos (int) --> Byte to move the pointer to
        """
        cls.stk_handler.seek(pos)

    @classmethod
    def getCompatibleFileTypes(cls) -> list:
        return ['stk', 'vol']


ALLOWED_COLUMNS_TYPES = ['String', 'Float', 'Integer', 'Boolean', 'Matrix',
                         'CsvList']
ADITIONAL_INFO_DISPLAY_COLUMN_LIST = ['_size', 'id']
EXCLUDED_COLUMNS = ['label', 'comment', 'creation', '_streamState']
PERMANENT_COLUMNS = ['id', 'enabled']
CLASS_OBJECT = 1
REPRESENTATIVE_OBJECT = 2
CLASS_ELEMENTS = 3
EXTENDED_COLUMN_NAME = 'stack'
ENABLED_COLUMN = 'enabled'
PROPERTIES_TABLE = 'Properties'
OBJECT_TABLE = 'objects'


class SqliteFile(IDAO):
    """  Class to manipulate Scipion Sqlite files. """

    def __init__(self, sqliteFile):
        self._names = []
        self._file = sqliteFile
        self._con = self.__loadDB(sqliteFile)
        self._con.row_factory = self._dictFactory
        self._tableCount = {}
        self._tables = {}
        self._labels = {}
        self._labelsTypes = {}
        self._aliases = {}
        self._columnsMap = {}
        self._extendedColumn = None
        self._tableWithAdditionalInfo = None
        self._objectsType = {}

    def __loadDB(self, sqliteFile):
        """Load a sqlite file"""
        try:
            return sqlite3.connect(f"file:{sqliteFile}?mode=ro", uri=True)
        except Exception as e:
            logger.error("The file could not be opened. Make sure the path is "
                         "correct: \n %s" % e)
            return None

    def hasExtendedColumn(self):
        """Return if the table need to extend a column. That column is used to
        renderer an image that is composed by other two columns"""
        return self._extendedColumn is not None

    def composeDataTables(self, tablesNames):
        """This method is used to generate a dictionary with the principal
           tables mapping the dependencies with other tables"""
        tablesNames = sorted(tablesNames)
        for tableName in tablesNames:
            divTable = tableName.split('_')
            if len(divTable) > 1:
                if divTable[-1].startswith('Class') and tableName not in self._tables:
                    objectTable = tableName.replace(divTable[-1], '') + 'Objects'
                    self._tables[objectTable] = tableName
                    self._names.append(objectTable)

        self._tables[PROPERTIES_TABLE] = PROPERTIES_TABLE
        self._names.append(PROPERTIES_TABLE)

    def composeObjectType(self):
        """Define the different objects types"""
        # General type defined into Properties table
        firstRow = self.getTableRow(PROPERTIES_TABLE, 0)
        objectType = firstRow['value']
        self._objectsType[self._aliases[OBJECT_TABLE]] = objectType

        for alias in self._aliases.values():
            objectTypeAux = alias.split('_')
            if len(objectTypeAux) == 2:
                sufix = 's' if objectTypeAux[0][-1] in "aeiouAEIOU" or objectTypeAux[0][-1] != 's' else ''
                objectType = 'SetOf%s%s' % (objectTypeAux[1], sufix)

                if objectTypeAux[1] not in self._objectsType:
                    self._objectsType[objectTypeAux[1]] = objectType

    def composeTableAlias(self, tableName):
        """Create an alias for the given table"""
        if tableName != PROPERTIES_TABLE:
            firstRow = self.getTableRow(tableName, 0)
            className = firstRow['class_name']
            if tableName.__contains__('_'):
                tableSplit = tableName.split('_')
                lenTableSplit = len(tableSplit)
                if lenTableSplit > 1:
                    alias = tableName.replace(tableSplit[-1],'') + className
            else:
                alias = className
        else:
            alias = PROPERTIES_TABLE
        return alias

    def getTableNames(self):
        """ Return all the table names found in the database. """
        initTime = time.time()
        if not self._names:
            self._tables = {OBJECT_TABLE: 'classes'}
            self._names = [OBJECT_TABLE]

            res = self._con.execute("SELECT name FROM sqlite_master WHERE type='table'")
            tablesNames = [row['name'] for row in res.fetchall()]
            self.composeDataTables(tablesNames)

            for tableName in self._names:
                # Getting the first row to identify the labels and theirs type
                firstRow = self.getTableRow(tableName, 0, classes=self._tables[tableName])
                self._labels[tableName] = [key for key in firstRow.keys() if key not in EXCLUDED_COLUMNS]
                alias = self.composeTableAlias(self._tables[tableName])
                self._aliases[tableName] = alias

                labelsTypes = []
                for key, value in firstRow.items():
                    if key not in EXCLUDED_COLUMNS:
                        labelsTypes.append(_guessType(value))
                self._labelsTypes[tableName] = labelsTypes

            self.composeObjectType()

        if len(self._tables) > 2: # Assuming that there are more tables than just Object and Properties
            self._tableWithAdditionalInfo = OBJECT_TABLE
        endTime = time.time()
        logging.info("GetTableNames: %f" % (endTime - initTime))
        return self._names

    def findColbyName(self, colNames, colName):
        """Return a column index given a column name"""
        for i, col in enumerate(colNames):
            if colName == col:
                return i
        return None

    def updateExtendColumn(self, table):
        """Find the columns that need to extend and keep the indexes"""
        tableName = table.getName()
        colNames = self._labels[tableName]
        indexCol = self.findColbyName(colNames, '_index')
        fileNameCol = self.findColbyName(colNames, '_filename')

        if indexCol and fileNameCol:
            logger.debug("The columns _index and _filename have been found. "
                         "We will proceed to create a new column with the "
                         "values of these columns.")
            self._extendedColumn = indexCol, fileNameCol
        else:
            indexCol = self.findColbyName(colNames, '_representative._index')
            fileNameCol = self.findColbyName(colNames,
                                                   '_representative._filename')
            if indexCol and fileNameCol:
                logger.debug("The columns _representative._index and "
                             "_representative._filename have been found. "
                             "We will proceed to create a new column with the "
                             "values of these columns.")
                self._extendedColumn = indexCol, fileNameCol

    def generateTableActions(self, table, objectManager):
        """Generate actions for a given table in order to create subsets"""
        if self.getScipionPort() and table.getName() != PROPERTIES_TABLE:
            alias = table.getAlias()
            labels = list(self._objectsType.keys())
            objectTypes = list(self._objectsType.values())
            aliasSplit = alias.split('_')
            if alias.startswith('Class') and len(aliasSplit) == 1:

                table.addAction(labels[0], lambda: self.createSubsetCallback(table, objectTypes[0], objectManager))
                table.addAction(labels[1], lambda: self.createSubsetCallback(table, objectTypes[1], objectManager))
                if alias == 'Class2D':
                    table.addAction('Averages', lambda: self.createSubsetCallback(table, 'SetOfAverages', objectManager))
                else:
                    table.addAction('Volumes', lambda: self.createSubsetCallback(table, 'SetOfVolumes',  objectManager))
            elif alias.startswith('Class'):
                    table.addAction(aliasSplit[1], lambda: self.createSubsetCallback(table, self._objectsType[aliasSplit[1]], objectManager))
            elif alias in self._objectsType and self._objectsType[alias].startswith('SetOf'):
                table.addAction(alias, lambda: self.createSubsetCallback(table, self._objectsType[alias], objectManager))

    def fillTable(self, table, objectManager):
        """Create the table structure (columns) and set the table alias"""
        initTime = time.time()
        tableName = table.getName()
        colNames = self._labels[tableName]
        self.updateExtendColumn(table)

        row = self.getTableRow(tableName, 0, classes=self._tables[tableName])
        if 'id' not in row:
            table.setHasColumnId(False)
        values = [value for key, value in row.items() if key not in EXCLUDED_COLUMNS]

        for index, colName in enumerate(colNames):

            isFileNameCol = self.hasExtendedColumn() and index == self._extendedColumn[1]

            if colName == ENABLED_COLUMN:
                renderer = BoolRenderer()
            elif isFileNameCol:
                renderer = StrRenderer()
            else:
                renderer = table.guessRenderer(str(values[index]))
                if isinstance(renderer, ImageRenderer):
                    imageExt = str(values[index]).split('.')[-1]
                    self.addExternalProgram(renderer, imageExt)

            newCol = Column(colName, renderer)
            newCol.setIsSorteable(True)
            newCol.setIsVisible(objectManager.isLabelVisible(colName))
            table.addColumn(newCol)

            if isFileNameCol:
                logger.debug("Creating an extended column: %s" % EXTENDED_COLUMN_NAME)
                imageExt = str(values[index]).split('.')[-1]
                renderer = ImageRenderer()
                self.addExternalProgram(renderer, imageExt)
                extraCol = Column(colName, renderer)
                extraCol.setIsVisible(newCol.isVisible())
                extraCol.setIsSorteable(False)
                table.addColumn(extraCol)
                newCol.setIsVisible(False)
                newCol.setName(EXTENDED_COLUMN_NAME)

        table.setAlias(self._aliases[tableName])
        self.generateTableActions(table, objectManager)
        endTime = time.time()
        logging.info("Create the table structure: %f" % (endTime - initTime))

    def addExternalProgram(self, renderer: ImageRenderer, imageExt: str):
        self.addChimera(renderer, imageExt)
        self.addImageJ(renderer)

    def addChimera(self, renderer: ImageRenderer, imageExt: str):
        chimeraPath = os.environ.get('CHIMERA_HOME', None)
        if chimeraPath is not None:
            if imageExt not in ['st', 'stk', 'mrcs']:
                icon = pw.findResource('chimera.png')

                def openChimeraCallback(imagePath):
                    path = imagePath.split('@')[-1]
                    program = os.path.join(chimeraPath, 'bin', 'ChimeraX')
                    cmd = program + ' "%s"' % path
                    Popen(cmd, shell=True, cwd=os.getcwd())

                renderer.addAction(Action('ChX', icon, 'Open with ChimeraX',
                                          openChimeraCallback))

    def addImageJ(self, renderer: ImageRenderer):
        imageJPath = os.environ.get('IMAGEJ_BINARY_PATH', None)
        if imageJPath is not None:
            icon = pw.findResource('Imagej.png')

            def openImageJCallback(imagePath):
                imageSplit = imagePath.split('@')
                selectedSlice = int(imageSplit[0])
                path = imageSplit[-1]
                program = os.path.join(imageJPath)
                macro = r"""
path = "%s";
open(path);
slice = %d; 
Stack.setSlice(slice);
                """ % (os.path.abspath(path), selectedSlice)

                with tempfile.NamedTemporaryFile(delete=False) as tempFile:
                    tempFile.write(macro.encode('utf-8'))
                    macroPath = tempFile.name

                cmd = program + ' -macro %s' % macroPath
                Popen(cmd, shell=True, cwd=os.getcwd())

            renderer.addAction(Action('IJ', icon, 'Open with ImageJ',
                                      openImageJCallback))

    def addImageViewer(self, renderer: ImageRenderer, imageExt: str):
        icon = pw.findResource('file_vol.png')

        def openImageViewerCallback(imagePath):
            pass

        if imageExt in ['mrc']:
            renderer.addAction(Action('IJ', icon, 'Open with ImageViewer',
                                      openImageViewerCallback))

    def fillPage(self, page, actualColumn=0, orderAsc=True):
        """
        Read the given table from the sqlite and fill the page(add rows)
        """
        initTime = time.time()
        tableName = page.getTable().getName()
        # moving to the first row of the page
        pageNumber = page.getPageNumber()
        pageSize = page.getPageSize()
        firstRow = pageNumber * pageSize - pageSize
        limit = pageSize

        column = self._labels[tableName][actualColumn]
        mode = 'ASC' if orderAsc else 'DESC'
        self.updateExtendColumn(page.getTable())

        for rowcount, row in enumerate(self.iterTable(tableName, start=firstRow, limit=limit,
                                       classes=self._tables[tableName],
                                       orderBy=column, mode=mode)):
            if row:
                values = [value for key, value in row.items() if key not in EXCLUDED_COLUMNS]
                if 'id' in row.keys():
                    id = row['id']
                else:
                    id = rowcount

                # Checking if exists an extended column
                if self.hasExtendedColumn() and tableName != PROPERTIES_TABLE:
                    values.insert(self._extendedColumn[1] + 1,
                                  str(values[self._extendedColumn[0]]) + '@' + values[self._extendedColumn[1]])
                page.addRow((int(id), values))
        endTime = time.time()
        logging.info("Fill the page: %f" % (endTime - initTime))

    def getRowsCount(self, tableName):
        """ Return the number of elements in the given table. """
        logger.debug("Reading the table %s" %tableName)
        if tableName == OBJECT_TABLE:
            size = int(self._con.execute(f"SELECT * FROM {PROPERTIES_TABLE} WHERE key='_size'").fetchall()[0]['value'])
            return size
        return self._con.execute(f"SELECT COUNT(*) FROM {tableName}").fetchone()['COUNT(*)']

    def getTableRowCount(self, tableName):
        if tableName not in self._tableCount:
            initTime = time.time()
            self._tableCount[tableName] = self.getRowsCount(tableName)
            endTime = time.time()
            logging.info("Table count: %f" % (endTime - initTime))
        return self._tableCount[tableName]

    def getSelectedRangeRowsIds(self, tableName, startRow, numberOfRows, column, reverse=True):
        """Return a range of rows starting at 'startRow' an amount
           of 'numberOfRows' """

        logger.debug("Reading the table %s and selected a range of rows %d - %d" % (tableName,startRow, numberOfRows + 1))
        mode = 'ASC' if reverse else 'DESC'
        col = self._getColumnMap(tableName, column)
        if col == None:
            col = column
        query = "SELECT id FROM %s ORDER BY %s %s LIMIT %d , %d" % (tableName, col, mode, startRow - 1, numberOfRows + 1)
        rowsList = self._con.execute(query).fetchall()
        rowsIds = [row['id'] for row in rowsList]
        return rowsIds

    def getColumnsValues(self, tableName, columns, xAxis, selection, limit,
                         useSelection, reverse=True):
        """Get the values of the selected columns in order to plot them"""

        logger.debug("Reading the table %s and selected some columns values...")
        cols = columns
        if xAxis and xAxis not in cols:
            cols.append(xAxis)
        columnNames = []
        for column in cols:
            col = self._getColumnMap(tableName, column) or column
            columnNames.append(col)

        columnNames = ", ".join(columnNames)

        col = self._getColumnMap(tableName, xAxis)
        if col is not None:
            xAxis = col

        mode = 'ASC' if reverse else 'DESC'
        orderBy = ' ORDER BY %s %s' % (xAxis, mode) if xAxis else ''
        limit = ' LIMIT %d' % limit if limit is not None else ''
        where = f" WHERE id in ({', '.join(map(str, selection.getSelection().keys()))})" if selection.getCount() > 1 and useSelection else ''

        query = "SELECT %s FROM %s %s %s %s" % (columnNames, tableName, where,
                                                 orderBy, limit)
        selectedColumns = self._con.execute(query).fetchall()

        columnsValues = {}

        firstValue = selectedColumns[0]
        for colName in columns:
            col = self._getColumnMap(tableName, colName) or colName
            columnsValues[colName] = [firstValue[col]]

        for pos, value in enumerate(selectedColumns):
            if pos > 0:
                for colName in columns:
                    col = self._getColumnMap(tableName, colName) or colName
                    columnsValues[colName].append((value[col]))

        return columnsValues

    def iterTable(self, tableName, **kwargs):
        """
        Method to iterate over the table's rows
        :param tableName: the name of the table
        :param kwargs:
                limit: integer value to limit the number of elements
                start: start from a given element
                classes: read column names from a 'classes' table
                orderBy: clause to sort given a column name
                mode: sort direction ASC or DESC
        """
        query = f"SELECT * FROM {tableName}"

        if 'mode' in kwargs:
            if 'orderBy' in kwargs:
                if kwargs['orderBy']:
                    column = self._getColumnMap(tableName, kwargs['orderBy'])
                    if not column:
                        column = kwargs['orderBy']

                    query += f" ORDER BY {column}"

            if kwargs['mode']:
                query += f" {kwargs['mode']}"

        if 'start' in kwargs and 'limit' not in kwargs:
            kwargs['limit'] = -1

        if 'limit' in kwargs:
            query += f" LIMIT {kwargs['limit']}"

        if 'start' in kwargs:
            query += f" OFFSET {kwargs['start']}"

        if 'classes' not in kwargs or kwargs['classes'] == PROPERTIES_TABLE:
            res = self._con.execute(query)
            while row := res.fetchone():
                yield row
        else:  # Mapping the column names and  including only the allowed columns
            self._columnsMap[tableName] = {row['column_name']: row['label_property']
                          for row in self.iterTable(kwargs['classes']) if row['class_name'] in ALLOWED_COLUMNS_TYPES}
            self._excludedColumns = {row['column_name']: row['label_property']
                          for row in self.iterTable(kwargs['classes']) if row['class_name'] not in ALLOWED_COLUMNS_TYPES}

            def _row_factory(cursor, row):
                fields = [column[0] for column in cursor.description]
                rowFact = {self._columnsMap[tableName].get(k, k): v for k, v in zip(fields, row) if k not in self._excludedColumns}
                return rowFact

            # Modify row factory to modify column names
            self._con.row_factory = _row_factory
            res = self._con.execute(query)
            while row := res.fetchone():
                yield row
            # Restore row factory
            self._con.row_factory = self._dictFactory

    def getTableAliases(self):
        """Return the tables aliases"""
        return self._aliases

    def _getColumnMap(self, tableName, column):
        """Get the column name that has been mapped"""
        if tableName in self._columnsMap:
            for key, value in self._columnsMap[tableName].items():
                if value == column:
                    return key
        return None

    def getTableRow(self, tableName, rowIndex, **kwargs):
        """ Get a given row by index. Extra args are passed to iterTable. """
        kwargs['start'] = rowIndex
        kwargs['limit'] = 1
        for row in self.iterTable(tableName, **kwargs):
            return row

    def getTableWithAdditionalInfo(self):
        """Return a tuple with the table that need to show additional info and
        the column that we need to show"""
        return self._tableWithAdditionalInfo, ADITIONAL_INFO_DISPLAY_COLUMN_LIST

    def createSubsetCallback(self, table: Table, objectType: str, objectManager):
        """Create a subset"""
        selection = table.getSelection().getSelection()
        tableName = table.getName()
        elementsCount = len(selection)
        if not elementsCount:
            elementsCount = self._tableCount[tableName]
        subsetName = objectManager.getGui().getSubsetName(objectType, elementsCount)
        if subsetName:
            format = '%Y%m%d%H%M%S'
            now = datetime.now()
            timestamp = now.strftime(format)
            path = 'Logs/selection_%s.txt' % timestamp
            self.writeSelection(table, path)
            path +="," # Always add a comma, it is expected by the user subset protocol
            if tableName != OBJECT_TABLE:
                path += tableName.split(OBJECT_TABLE)[0]

        self.sendSubsetCreationMessage(path, objectType, subsetName)

    def sendSubsetCreationMessage(self, selectionFile, outputClassName, label):

        import socket

        # Create a client socket
        clientSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM);

        # Connect to the local server where Scipion GUI is listening
        clientSocket.connect(("127.0.0.1", int(self.getScipionPort())));

        # We should create this message:
        # run protocol ProtUserSubSet inputObject=380 sqliteFile='...','' outputClassName=SetOfTiltSeries other='' label='create subset'
        data = f"run protocol ProtUserSubSet inputObject={self.getScipionObjectId()} " \
               f"sqliteFile='{selectionFile}' outputClassName='{outputClassName}' other='' label='{label}'"

        clientSocket.send(data.encode());


    def getScipionPort(self):
        """ Returns Scipion port or None if not in the environment"""
        return os.getenv(SCIPION_PORT, '1300')

    def getScipionObjectId(self):
        """ Returns Scipion object id"""
        return os.getenv(SCIPION_OBJECT_ID)

    def writeSelection(self, table: Table, path):
        """ Create a file with the selected rows ids"""
        tableName = table.getName()
        rowsIds = table.getSelection().getSelection().keys()
        if not rowsIds:
            rowsIds = [i+1 for i in range(self._tableCount[tableName])]
        try:
            with open(path, 'w') as file:
                for rowId in rowsIds:
                    file.write(str(rowId) + ' ')
                file.close()
            logger.debug(f"The file: {path} was created correctly.")
        except Exception as e:
            logger.error(f"Error creating the file: {e}")

    def close(self):
        if getattr(self, '_con', None):
            self._con.close()
            self._con = None

    def _dictFactory(self, cursor, row):
        fields = [column[0] for column in cursor.description]
        return {key: value for key, value in zip(fields, row)}

    @classmethod
    def getCompatibleFileTypes(cls):
        """Return a list of compatible extension of files"""
        logger.debug("Selected SqliteFile DAO")
        return ['sqlite']

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()


# --------- Helper functions  ------------------------
def _guessType(strValue):
    if strValue is None:
        return str('None')
    try:
        int(strValue)
        return int
    except ValueError:
        try:
            float(strValue)
            return float
        except ValueError:
            return str


def extendMDViewer(om:metadataviewer.model.ObjectManager):
    """ Function to extend the object manager with DAOs and readers"""
    om.registerDAO(SqliteFile)
    om.registerReader(MRCImageReader)
    om.registerReader(STKImageReader)


# **************************************************************************
# *
# * Authors:     Airen Zaldivar Peraza (azaldivar@cnb.csic.es) [1]
# *              J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import enum
import os
from datetime import datetime
from collections import OrderedDict

import pyworkflow.object as pwobj
import pyworkflow.protocol as pwprot
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

import pwem.constants as emcts
import pwem.objects as emobj
from pwem.protocols import EMProtocol


class ProtParticles(EMProtocol):
    pass


class ProtProcessParticles(ProtParticles):
    """ This class will serve as a base for all protocol
    that performs some operation on Particles (i.e. filters, mask, resize, etc)
    It is mainly defined by an inputParticles and outputParticles.
    """

    def _defineParams(self, form):
        form.addSection(label=pwutils.Message.LABEL_INPUT)

        form.addParam('inputParticles', params.PointerParam,
                      pointerClass='SetOfParticles',
                      label=pwutils.Message.LABEL_INPUT_PART, important=True)
        # Hook that should be implemented in subclasses
        self._defineProcessParams(form)

        __threads, __mpi = self._getDefaultParallel()

        form.addParallelSection(threads=__threads, mpi=__mpi)

    def _defineProcessParams(self, form):
        """ This method should be implemented by subclasses
        to add other parameter relatives to the specific operation."""
        pass

    def _getDefaultParallel(self):
        """ Return the default value for thread and MPI
        for the parallel section definition.
        """
        return 0, 0


class ProtFilterParticles(ProtProcessParticles):
    """ Base class for filters on particles of type ProtPreprocessParticles.
    """
    pass


class ProtOperateParticles(ProtProcessParticles):
    """ Base class for operations on particles of type ProtPreprocessParticles.
    """

    def __init__(self, **args):
        ProtProcessParticles.__init__(self, **args)


class ProtMaskParticles(ProtProcessParticles):
    """ This is the base for the branch of mask, 
    between the ProtPreprocessParticles """
    pass


# Micrograph type constants for particle extraction
SAME_AS_PICKING = 0
OTHER = 1


class ProtExtractParticlesOutput(enum.Enum):
    """Predefined outputs for particle extraction protocols"""
    outputParticles = emobj.SetOfParticles


# noinspection SqlDialectInspection
class ProtExtractParticles(ProtParticles):
    """ Base class for all extract-particles protocols.
     This class will take care of the streaming functionality and
     derived classes should mainly overwrite the '_extractMicrograph' function.
     """
    _possibleOutputs = ProtExtractParticlesOutput
    # --------------------------- DEFINE param functions ------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputCoordinates', params.PointerParam,
                      pointerClass='SetOfCoordinates',
                      important=True,
                      label="Input coordinates",
                      help='Select the SetOfCoordinates ')

        # The name for the following param is because historical reasons
        # now it should be named better 'micsSource' rather than
        # 'downsampleType', but this could make inconsistent previous executions
        # of this protocols, we will keep the name
        form.addParam('downsampleType', params.EnumParam,
                      choices=['same as picking', 'other'],
                      default=0, important=True,
                      display=params.EnumParam.DISPLAY_HLIST,
                      label='Micrographs source',
                      help='By default the particles will be extracted '
                           'from the micrographs used in the picking '
                           'step ( _same as picking_ option ). \n'
                           'If you select _other_ option, you must provide '
                           'a different set of micrographs to extract from. \n'
                           '*Note*: In the _other_ case, ensure that provided '
                           'micrographs and coordinates are related '
                           'by micName or by micId. Difference in pixel size '
                           'will be handled automatically.')

        form.addParam('inputMicrographs', params.PointerParam, allowsNull=True,
                      pointerClass='SetOfMicrographs',
                      condition='downsampleType != %s' % SAME_AS_PICKING,
                      important=True, label='Input micrographs',
                      help='Select the SetOfMicrographs from which to extract.')

        form.addParam('ctfRelations', params.RelationParam, allowsNull=True,
                      condition='inputCoordinates is not None',
                      relationName=emcts.RELATION_CTF,
                      attributeName='getInputMicrographs',
                      label='CTF estimation',
                      help='Choose some CTF estimation related to input '
                           'micrographs. \n CTF estimation is needed if you '
                           'want to do phase flipping or you want to '
                           'associate CTF information to the particles.')

        self._definePreprocessParams(form)

    def _definePreprocessParams(self, form):
        """ Should be implemented in sub-classes to define some
        specific parameters """
        pass

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        # Let's load input data for the already existing micrographs
        # before the streaming
        self.debug(">>> _insertAllSteps ")
        pwutils.makeFilePath(self._getAllDone())
        self.micDict = OrderedDict()
        self.coordDict = {}

        micDict = self._loadInputList()

        self.initialIds = self._insertInitialSteps()

        pickMicIds = self._insertNewMicsSteps(micDict.values())

        self._insertFinalSteps(pickMicIds)

    def _insertInitialSteps(self):
        """ Override this function to insert some steps before the
        extract micrograph steps.
        Should return a list of ids of the initial steps. """
        return []

    def _insertNewMicsSteps(self, inputMics):
        """ Insert steps to process new mics (from streaming)
        Params:
            inputMics: input mics set to be check
        """
        return self._insertNewMics(inputMics,
                                   lambda mic: mic.getMicName(),
                                   self._insertExtractMicrographStep,
                                   self._insertExtractMicrographListStep,
                                   *self._getExtractArgs())

    def _insertFinalSteps(self, micSteps):
        """ Override this function to insert some steps after the
        extraction of micrograph steps.
        Receive the list of step ids of the picking steps. """
        self._insertFunctionStep('createOutputStep',
                                 prerequisites=micSteps, wait=True)

    def _insertExtractMicrographStep(self, mic, prerequisites, *args):
        """ Basic method to insert a picking step for a given micrograph. """
        micStepId = self._insertFunctionStep('extractMicrographStep',
                                             mic.getMicName(), *args,
                                             prerequisites=prerequisites)
        return micStepId

    # -------------------------- STEPS functions ------------------------------

    def extractMicrographStep(self, micKey, *args):
        """ Step function that will be common for all extraction protocols.
        It will take an id and will grab the micrograph from a micDict map.
        The micrograph will be passed as input to the _extractMicrograph
        function.
        """
        # Retrieve the corresponding micrograph with this key and the
        # associated list of coordinates
        mic = self.micDict[micKey]

        micDoneFn = self._getMicDone(mic)
        micFn = mic.getFileName()

        if self.isContinued() and os.path.exists(micDoneFn):
            self.info("Skipping micrograph: %s, seems to be done" % micFn)
            return

        coordList = self.coordDict[mic.getObjId()]
        self._convertCoordinates(mic, coordList)

        # Clean old finished files
        pwutils.cleanPath(micDoneFn)

        self.info("Extracting micrograph: %s " % micFn)
        self._extractMicrograph(mic, *args)

        # Mark this mic as finished
        open(micDoneFn, 'w').close()

    def _extractMicrograph(self, mic, *args):
        """ This function should be implemented by subclasses in order
        to pick the given micrograph. """
        pass

    # ---------- Methods to extract many micrographs at once ------------------

    def _insertExtractMicrographListStep(self, micList, prerequisites, *args):
        """ Basic method to insert a picking step for a given micrograph. """
        return self._insertFunctionStep('extractMicrographListStep',
                                        [mic.getMicName() for mic in micList],
                                        *args, prerequisites=prerequisites)

    def extractMicrographListStep(self, micKeyList, *args):
        micList = []

        for micName in micKeyList:
            mic = self.micDict[micName]
            micDoneFn = self._getMicDone(mic)
            micFn = mic.getFileName()
            if self.isContinued() and os.path.exists(micDoneFn):
                self.info("Skipping micrograph: %s, seems to be done" % micFn)

            else:
                # Clean old finished files
                pwutils.cleanPath(micDoneFn)
                self.info("Extracting micrograph: %s " % micFn)
                micList.append(mic)

        self._extractMicrographList(micList, *args)

        for mic in micList:
            # Mark this mic as finished
            open(self._getMicDone(mic), 'w').close()

    def _extractMicrographList(self, micList, *args):
        """ Extract more than one micrograph at once.
        Here the default implementation is to iterate through the list and
        call the single extract, but it could be re-implemented on each
        subclass to provide a more efficient implementation.
        """
        for mic in micList:
            self._extractMicrograph(mic, *args)

    # --------------------------- UTILS functions -----------------------------
    def _convertCoordinates(self, mic, coordList):
        """ This function should be implemented by subclasses. """
        pass

    def _getExtractArgs(self):
        """ Should be implemented in sub-classes to define the argument
        list that should be passed to the extract step function.
        """
        return []

    def _micsOther(self):
        """ Return True if other micrographs are used for extract.
        Should be implemented in derived classes.
        """
        return False

    def _useCTF(self):
        """ Return True if a SetOfCTF is associated with the extracted
        particles.
         Should be implemented in derived classes.
        """
        return False

    # ------ Methods for Streaming extraction --------------
    def _isStreamClosed(self):
        return self.micsClosed and self.ctfsClosed and self.coordsClosed

    def _stepsCheck(self):
        # To allow streaming picking we need to detect:
        #   1) new micrographs ready to be picked
        #   2) new output coordinates that have been produced and add then
        #      to the output set.
        self._checkNewInput()
        self._checkNewOutput()

    def _loadInputList(self):
        """ Load the input set of micrographs and create a dictionary.
        The dictionary self.micDict, will contain all the micrographs
        from which particles have been extracted and also ones that
        are ready to be extracted.
        For creating this dictionary, we need to inspect:
        1) Input micrographs coming from coordinates, to know that a given
        micrographs has been picked.
        2) Micrographs to be extracted (in case it is different from 1)
        3) New computed CTF (in case it is associated with the particles)
        """

        def _loadSet(inputSet, SetClass, getKeyFunc):
            setFn = inputSet.getFileName()
            self.debug("Loading input db: %s" % setFn)
            updatedSet = SetClass(filename=setFn)
            updatedSet.loadAllProperties()
            newItemDict = OrderedDict()
            for item in updatedSet:
                micKey = getKeyFunc(item)
                if micKey not in self.micDict:
                    newItemDict[micKey] = item.clone()
            streamClosed = updatedSet.isStreamClosed()
            updatedSet.close()
            self.debug("Closed db.")
            return newItemDict, streamClosed

        def _loadMics(micSet):
            return _loadSet(micSet, emobj.SetOfMicrographs,
                            lambda mic: mic.getMicName())

        def _loadCTFs(ctfSet):
            return _loadSet(ctfSet, emobj.SetOfCTF,
                            lambda ctf: ctf.getMicrograph().getMicName())

        # Load new micrographs coming from the coordinates
        self.debug("Loading Mics from Coords.")
        coordMics = self.inputCoordinates.get().getMicrographs()
        micDict, self.micsClosed = _loadMics(coordMics)
        # If we are extracting from other micrographs, then we will use
        # the other micrographs and filter those that
        if self._micsOther():
            self.debug("Loading other Mics.")
            oMicDict, oMicClosed = _loadMics(self.inputMicrographs.get())
            self.micsClosed = self.micsClosed and oMicClosed
            micDictNew = {}
            for micKey, mic in micDict.items():
                if micKey in oMicDict:
                    oMic = oMicDict[micKey]
                    # Let's fix the id in case it does not correspond
                    # we want to have the id coming from the coordinates
                    # to match each coordinate to its micrograph
                    oMic.copyObjId(mic)
                    micDictNew[micKey] = oMic
            micDict = micDictNew

        self.debug("Mics are closed? %s" % self.micsClosed)
        if self._useCTF():
            self.debug("Loading CTFs.")
            ctfDict, ctfClosed = _loadCTFs(self.ctfRelations.get())
            micDictNew = {}
            for micKey, mic in micDict.items():
                if micKey in ctfDict:
                    mic.setCTF(ctfDict[micKey])
                    micDictNew[micKey] = mic
            micDict = micDictNew

        # if not use CTF, self.ctfsClosed is True
        self.ctfsClosed = ctfClosed if self._useCTF() else True
        self.debug("CTFs are closed? %s" % self.ctfsClosed)

        self.debug("Loading Coords.")
        # Now load the coordinates for the newly detected micrographs. If
        # micrographs does not have coordinates, is not processed.
        micDict = self._loadInputCoords(micDict)

        # Store this value to be used when inserting new steps and batch mode
        self.streamClosed = self._isStreamClosed()

        return micDict

    def _loadInputCoords(self, micDict):
        """ Load coordinates from the input streaming.
        """
        # TODO: this takes for ever if you are NOT
        # doing streaming and have several thousands of mics
        # so I add a counter to keep the user entertained
        import sys
        a = datetime.now()
        counter = 1

        coordsFn = self.getCoords().getFileName()
        self.debug("Loading input db: %s" % coordsFn)
        coordSet = emobj.SetOfCoordinates(filename=coordsFn)
        # FIXME: Temporary to avoid loadAllPropertiesFail
        coordSet._xmippMd = pwobj.String()
        coordSet.loadAllProperties()

        micList = dict()  # To store a dictionary with mics with coordinates

        for micKey, mic in micDict.items():
            if counter % 50 == 0:
                b = datetime.now()
                print(b - a, 'reading coordinates for mic number', "%06d" % counter)
                sys.stdout.flush()  # force buffer to print
            counter += 1

            micId = mic.getObjId()
            coordList = []

            self.debug("Loading coords for mic: %s (%s)" % (micId, micKey))
            for coord in coordSet.iterItems(where='_micId=%s' % micId):
                # TODO: Check performance penalty of using this clone
                coordList.append(coord.clone())
            self.debug("Coords found: %s" % len(coordList))

            if coordList:
                self.coordDict[micId] = coordList
                micList[micKey] = mic

        self.coordsClosed = coordSet.isStreamClosed()
        coordSet.close()
        self.debug("Coords are closed? %s" % self.coordsClosed)
        self.debug("Closed db.")

        return micList

    def _checkNewInput(self):
        self.debug(">>> _checkNewInput ")

        def _modificationTime():
            """ Check the last modification time of any of the three possible
             input files. """
            items = [self.inputCoordinates.get()]

            if self._micsOther():
                items.append(self.inputMicrographs.get())
            else:
                items.append(self.inputCoordinates.get().getMicrographs())

            if self._useCTF():
                items.append(self.ctfRelations.get())

            def _mTime(fn):
                return datetime.fromtimestamp(os.path.getmtime(fn))

            return max([_mTime(i.getFileName()) for i in items])

        mTime = _modificationTime()
        now = datetime.now()
        self.lastCheck = getattr(self, 'lastCheck', now)
        self.debug('Last check: %s, modification: %s'
                   % (pwutils.prettyTime(self.lastCheck),
                      pwutils.prettyTime(mTime)))
        # If the input micrographs.sqlite have not changed since our last check,
        # it does not make sense to check for new input data, but we must
        # check if sets are closed.
        self.debug("self.lastCheck > mTime %s , hasattr(self, 'micDict') %s"
                   % (self.lastCheck > mTime, hasattr(self, 'micDict')))
        if self.lastCheck > mTime and hasattr(self, 'micDict'):
            return None

        # Open input micrographs.sqlite and close it as soon as possible
        newMics = self._loadInputList()

        self.lastCheck = now
        outputStep = self._getFirstJoinStep()

        if newMics:
            fDeps = self._insertNewMicsSteps(newMics.values())
            if outputStep is not None:
                outputStep.addPrerequisites(*fDeps)
            self.updateSteps()

    def _checkNewOutput(self):
        if getattr(self, 'finished', False):
            return

        # Load previously done items (from text file)
        doneList = self._readDoneList()
        # Check for newly done items
        newDone = [m for m in self.micDict.values()
                   if m.getObjId() not in doneList and self._isMicDone(m)]

        # Update the file with the newly done mics
        # or exit from the function if no new done mics
        inputLen = len(self.micDict)
        self.debug('_checkNewOutput: ')
        self.debug('   input: %s, doneList: %s, newDone: %s'
                   % (inputLen, len(doneList), len(newDone)))

        firstTime = len(doneList) == 0
        allDone = len(doneList) + len(newDone)
        # We have finished when there is not more input mics (stream closed)
        # and the number of processed mics is equal to the number of inputs
        streamClosed = self._isStreamClosed()
        self.finished = streamClosed and allDone == inputLen
        self.debug(' is finished? %s ' % self.finished)
        self.debug(' is stream closed? %s ' % streamClosed)
        streamMode = pwobj.Set.STREAM_CLOSED if self.finished else pwobj.Set.STREAM_OPEN

        if newDone:
            self._updateOutputPartSet(newDone, streamMode)
            self._writeDoneList(newDone)
        elif not self.finished:
            # If we are not finished and no new output have been produced
            # it does not make sense to proceed and updated the outputs
            # so we exit from the function here

            # Maybe it would be good idea to take a snap to avoid
            # so much IO if this protocol does not have much to do now
            if allDone == len(self.micDict):
                self._streamingSleepOnWait()

            return

        self.debug('   finished: %s ' % self.finished)
        self.debug('        self.streamClosed (%s) AND' % streamClosed)
        self.debug('        allDone (%s) == len(self.listOfMics (%s)'
                   % (allDone, inputLen))
        self.debug('   streamMode: %s' % streamMode)

        if self.finished:  # Unlock createOutputStep if finished all jobs

            # Close the output set
            self._updateOutputPartSet([], pwobj.Set.STREAM_CLOSED)
            outputStep = self._getFirstJoinStep()
            if outputStep and outputStep.isWaiting():
                outputStep.setStatus(pwprot.STATUS_NEW)

    def readPartsFromMics(self, micDoneList, outputParts):
        """ This method should be implemented in subclasses to read
        the particles from a given list of micrographs.
        """
        pass

    def _updateOutputPartSet(self, micList, streamMode):
        outputName = ProtExtractParticlesOutput.outputParticles.name
        outputParts = getattr(self, outputName, None)
        firstTime = True

        if outputParts is None:
            inputMics = self.getInputMicrographs()
            outputParts = self._createSetOfParticles()
            outputParts.copyInfo(inputMics)
            outputParts.setCoordinates(self.inputCoordinates)

            if self.getAttributeValue('doFlip', False):
                outputParts.setIsPhaseFlipped(not inputMics.isPhaseFlipped())

            outputParts.setSamplingRate(self._getNewSampling())
            outputParts.setHasCTF(self._useCTF())
        else:
            firstTime = False
            outputParts.enableAppend()

        self.readPartsFromMics(micList, outputParts)
        self._updateOutputSet(outputName, outputParts, streamMode)

        if firstTime:
            # self._storeMethodsInfo(fnImages)
            self._defineSourceRelation(self.inputCoordinates, outputParts)
            if self._useCTF():
                self._defineSourceRelation(self.ctfRelations, outputParts)
            if self._micsOther():
                self._defineSourceRelation(self.inputMicrographs, outputParts)

    def _getMicDone(self, mic):
        return self._getExtraPath('DONE', 'mic_%06d.TXT' % mic.getObjId())

    def _isMicDone(self, mic):
        """ A mic is done if the marker file exists. """
        return os.path.exists(self._getMicDone(mic))

    def _getAllDone(self):
        return self._getExtraPath('DONE', 'all.TXT')

    def _readDoneList(self):
        """ Read from a text file the id's of the items that have been done. """
        doneFile = self._getAllDone()
        doneList = []
        # Check what items have been previously done
        if os.path.exists(doneFile):
            with open(doneFile) as f:
                doneList += [int(line.strip()) for line in f]

        return doneList

    def _writeDoneList(self, micList):
        """ Write to a text file the items that have been done. """
        doneFile = self._getAllDone()

        if not os.path.exists(doneFile):
            pwutils.makeFilePath(doneFile)

        with open(doneFile, 'a') as f:
            for mic in micList:
                f.write('%d\n' % mic.getObjId())

    def _getFirstJoinStepName(self):
        # This function will be used for streaming, to check which is
        # the first function that need to wait for all micrographs
        # to have completed, this can be overwritten in subclasses
        # (eg in Xmipp 'sortPSDStep')
        return 'createOutputStep'

    def _getFirstJoinStep(self):
        for s in self._steps:
            if s.funcName == self._getFirstJoinStepName():
                return s
        return None

    def createOutputStep(self):
        pass  # Nothing to do now
        # self._createOutput(self._getExtraPath())


class ProtExtractParticlesPair(ProtParticles):
    """ Base class for all extract-particles pairs protocols. Until now,
    this protocols is not in streaming mode.
     """



"""
Created on May 20, 2013

@author: laura
"""

from glob import iglob
import sqlite3
from unittest import TestCase

import numpy as np

from pyworkflow import SCIPION_DEBUG_NOCLEAN
from pyworkflow.tests import *
import pyworkflow.utils as pwutils

import pwem.objects as emobj
from pwem import emlib
from pwem.emlib import metadata as md
import pwem.protocols as emprot
import pyworkflow.tests as pwtests
from pwem.convert import SequenceHandler

# set to true if you want to check how fast is the access to
# the database
SPEEDTEST = True


def getIndex(fileName):
    """ Returns list of indexes for the table 'objects'
    in the database fileName"""
    conn = sqlite3.connect(fileName)
    cur = conn.cursor()
    cur.execute("PRAGMA index_list(objects);")
    rows = cur.fetchall()
    conn.close()
    return rows  # list with index names


def dropIndex(fileName, indexName):
    """ Drop index called indexName for table objects
    in database fileName"""
    conn = sqlite3.connect(fileName)
    cur = conn.cursor()
    cur.execute("DROP INDEX %s" % indexName)
    conn.close()


def createDummyProtocol(projName):
    """create a dummy protocol, returns protocol object"""
    proj = Manager().createProject(projName)
    os.chdir(proj.path)

    from pwem.protocols import EMProtocol

    prot = proj.newProtocol(EMProtocol)
    prot.setObjLabel('dummy protocol')
    try:
        proj.launchProtocol(prot)
    except Exception as e:
        logger.error("Can't launch EMProtocol", exc_info=e)
    return prot


class TestFSC(unittest.TestCase):
    _labels = [SMALL, WEEKLY]

    def testIO(self):
        """Test basic FSC object"""
        xList = [0.00, 0.05, 0.10, 0.15, 0.2]
        yList = [1.00, 0.95, 0.90, 0.85, 0.2]
        fsc = emobj.FSC()
        fsc.setData(xList, yList)
        x, y = fsc.getData()
        self.assertEqual(xList, x)
        self.assertEqual(yList, y)

    def testMd(self):
        """test create FSC from metadata"""
        xList = [0.00, 0.05, 0.10, 0.15, 0.2]
        yList = [1.00, 0.95, 0.90, 0.85, 0.2]
        md1 = emlib.MetaData()
        for freq, fscValue in zip(xList, yList):
            id = md1.addObject()
            md1.setValue(emlib.MDL_RESOLUTION_FREQ, freq, id)
            md1.setValue(emlib.MDL_RESOLUTION_FRC, fscValue, id)
        fsc = emobj.FSC()
        fsc.loadFromMd(md1, emlib.MDL_RESOLUTION_FREQ,
                       emlib.MDL_RESOLUTION_FRC)
        x, y = fsc.getData()
        self.assertEqual(xList, x)
        self.assertEqual(yList, y)


class TestImage(unittest.TestCase):
    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('xmipp_tutorial')
        cls.mic1 = cls.dataset.getFile('mic1')

    def testLocation(self):
        fn = self.mic1
        mic = emobj.Micrograph()
        mic.setFileName(fn)

        # Check that setFileName-getFileName is working properly
        self.assertEqual(fn, mic.getFileName())

        # Check the location is accepted from constructor
        mic2 = emobj.Micrograph(fn)
        self.assertEqual(fn, mic2.getFileName())

        volStk = '/data/current/volumes/all_volumes.stk'
        vol1 = emobj.Volume((1, volStk))
        self.assertEqual(1, vol1.getIndex())
        self.assertEqual(volStk, vol1.getFileName())

        self.assertEqual('all_volumes.stk', vol1.getBaseName())

    def testOrigin(self):
        fn = self.mic1
        mic = emobj.Micrograph()
        mic.setFileName(fn)
        mic.setSamplingRate(1.)

        referenceT = emobj.Transform()
        referenceT.setShifts(-4608., -4720., 1)
        mreferenceT = referenceT.getMatrix()

        t = mic.getOrigin(True)
        mt = t.getMatrix()
        for i in range(4):
            for j in range(4):
                self.assertAlmostEquals(mt[i][j], mreferenceT[i][j], delta=0.5)


class TestImageHandler(unittest.TestCase):
    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('xmipp_tutorial')
        cls.dsFormat = DataSet.getDataSet('movies')
        cls.dsRelion = DataSet.getDataSet('relion_tutorial')

    def testExistLocation(self):
        volFn = self.dataset.getFile('volumes/volume_1_iter_002.mrc')

        ih = emlib.image.ImageHandler()
        # Test the volume filename exists
        self.assertTrue(ih.existsLocation(volFn))
        # Test missing filename
        self.assertFalse(ih.existsLocation(volFn.replace('.mrc', '_fake.mrc')))
        # Test the :mrc is append when used as volume
        newFn = ih.getVolFileName(volFn)
        self.assertEqual(newFn, volFn + ":mrc")
        # Test that the new filename still exists even with the :mrc suffix
        self.assertTrue(ih.existsLocation(newFn))

    def testGetDimensions(self):
        volLabel = ':mrc'
        movieLabel = ':mrcs'
        ih = emlib.image.ImageHandler()

        # MICROGRAPH
        expectedSize_Mic = [9216, 9441, 1, 1]
        micFn = self.dataset.getFile('micrographs/BPV_1386.mrc')
        self.assertTrue(ih.existsLocation(micFn))
        X, Y, Z, N = ih.getDimensions(micFn)
        self.assertEqual([X, Y, Z, N], expectedSize_Mic)
        # Labelled as movie
        X, Y, Z, N = ih.getDimensions(micFn + movieLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mic)
        # Labelled as volume
        X, Y, Z, N = ih.getDimensions(micFn + volLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mic)

        # MOVIE .MRC
        expectedSize_Mov = [4096, 4096, 1, 7]
        expectedSize_Vol = [4096, 4096, 7, 1]
        movFn = self.dsFormat.getFile('qbeta/qbeta.mrc')  # Not labelled, so treated as volume
        self.assertTrue(ih.existsLocation(movFn))
        X, Y, Z, N = ih.getDimensions(movFn)
        self.assertEqual([X, Y, Z, N], expectedSize_Vol)
        # Labelled as movie
        X, Y, Z, N = ih.getDimensions(movFn + movieLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mov)
        # Labelled as volume
        X, Y, Z, N = ih.getDimensions(movFn + volLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Vol)

        # MOVIE .MRCS
        expectedSize_Mov = [1950, 1950, 1, 16]
        expectedSize_Vol = [1950, 1950, 16, 1]
        movFn = self.dsFormat.getFile('Falcon_2012_06_12-16_55_40_0_movie.mrcs')
        self.assertTrue(ih.existsLocation(movFn))
        X, Y, Z, N = ih.getDimensions(movFn)
        self.assertEqual([X, Y, Z, N], expectedSize_Mov)
        # Labelled as movie
        X, Y, Z, N = ih.getDimensions(movFn + movieLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mov)
        # Labelled as volume
        X, Y, Z, N = ih.getDimensions(movFn + volLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Vol)

        # VOLUME
        expectedSize_Mov = [64, 64, 1, 64]
        expectedSize_Vol = [64, 64, 64, 1]
        volFn = self.dataset.getFile('volumes/volume_1_iter_002.mrc')
        self.assertTrue(ih.existsLocation(volFn))
        X, Y, Z, N = ih.getDimensions(volFn)
        self.assertEqual([X, Y, Z, N], expectedSize_Vol)
        # Labelled as movie
        X, Y, Z, N = ih.getDimensions(volFn + movieLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mov)
        # Labelled as volume
        X, Y, Z, N = ih.getDimensions(volFn + volLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Vol)

        # STACK OF VOLUMES
        expectedSize_Mov = [60, 60, 1, 180]
        expectedSize_Svol = [60, 60, 60, 3]
        sVolFn = self.dsRelion.getFile('import/case2/relion_volumes.mrc')
        self.assertTrue(ih.existsLocation(sVolFn))
        X, Y, Z, N = ih.getDimensions(sVolFn)
        self.assertEqual([X, Y, Z, N], expectedSize_Svol)
        # Labelled as movie
        X, Y, Z, N = ih.getDimensions(sVolFn + movieLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Mov)
        # Labelled as volume
        X, Y, Z, N = ih.getDimensions(sVolFn + volLabel)
        self.assertEqual([X, Y, Z, N], expectedSize_Svol)

        # eer format (In case we want to cancel the warning)
        import logging
        tifffileLogger = logging.getLogger("tifffile.tifffile")
        tifffileLogger.disabled = True

        eerFile = self.dsFormat.getFile("eer")
        logger.info("Reading dimension for %s" % eerFile )
        X, Y, Z, N = ih.getDimensions(eerFile)
        self.assertEqual([X, Y, Z, N], [4096, 4096, 1, 567])

    def test_convertMicrographs(self):
        """ Convert micrographs to different formats.
         EMAN2 required for .img
        """
        micFn = self.dataset.getFile('micrographs/BPV_1386.mrc')
        outSuffix = pwutils.replaceBaseExt(micFn, 'img')
        ih = emlib.image.ImageHandler()

        outFn = join('/tmp', outSuffix)
        logger.info("Converting: \n%s -> %s" % (micFn, outFn))

        ih.convert(micFn, outFn)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)

        pwutils.cleanPath(outFn)
        pwutils.cleanPath(outFn.replace('.img', '.hed'))

    def test_readDM4(self):
        """ Check we can read dm4 files (using EMAN)
        """
        micFn = self.dsFormat.getFile('SuperRef_c3-adp-se-xyz-0228_001.dm4')

        ih = emlib.image.ImageHandler()
        # Check that we can read the dimensions of the dm4 file:
        EXPECTED_SIZE = (7676, 7420, 1, 1)
        self.assertEqual(ih.getDimensions(micFn), EXPECTED_SIZE)

        # We could even convert to an mrc file:
        outSuffix = pwutils.replaceBaseExt(micFn, 'mrc')

        outFn = join('/tmp', outSuffix)
        logger.info("Converting: \n%s -> %s" % (micFn, outFn))

        ih.convert(micFn, outFn)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        # Check dimensions are still the same:
        self.assertEqual(ih.getDimensions(outFn), EXPECTED_SIZE)

        # Clean up tmp files
        pwutils.cleanPath(outFn)

    def test_readCompressedTIF(self):
        """ Check we can read tif files
        """
        micFn = self.dsFormat.getFile('c3-adp-se-xyz-0228_200.tif')

        logger.info("Reading and converting %s" % micFn)
        ih = emlib.image.ImageHandler()
        # Check that we can read the dimensions of the dm4 file:
        EXPECTED_SIZE = (7676, 7420, 1, 38)
        self.assertEqual(ih.getDimensions(micFn), EXPECTED_SIZE)

        # We could even convert to an mrc file:
        outSuffix = pwutils.replaceBaseExt(micFn, 'mrc')

        outFn = join('/tmp', outSuffix)
        logger.info("Converting: \n%s -> %s" % ((1, micFn), outFn))

        ih.convert((1, micFn), outFn)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), (7676, 7420, 1, 1))

        # Clean up tmp files
        pwutils.cleanPath(outFn)

    def test_convertMovie(self):
        """Check movie conversion"""
        movFn = self.dsFormat.getFile('qbeta/qbeta.mrc') + ":mrcs"

        ih = emlib.image.ImageHandler()
        # Check that we can read the dimensions of the dm4 file:
        EXPECTED_SIZE = (4096, 4096, 1, 7)
        EXPECTED_DT = emlib.DT_USHORT

        self.assertEqual(ih.getDimensions(movFn), EXPECTED_SIZE)
        self.assertEqual(ih.getDataType(movFn), EXPECTED_DT)

        outFn = join('/tmp/qbeta_converted.mrcs')

        ih.convertStack(movFn, outFn, 2, 6)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), (4096, 4096, 1, 5))
        self.assertEqual(ih.getDataType(outFn), EXPECTED_DT)

        if pwutils.envVarOn(SCIPION_DEBUG_NOCLEAN):
            logger.info("Not cleaning output movie: ", outFn)
        else:
            pwutils.cleanPath(outFn)

    def test_truncateMask(self):
        ih = emlib.image.ImageHandler()

        maskFn = self.dataset.getFile('masks/mask.vol')
        outFn = join('/tmp', 'mask.vol')

        ih.truncateMask(maskFn, outFn, newDim=128)
        EXPECTED_SIZE = (128, 128, 128, 1)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), EXPECTED_SIZE)

        pwutils.cleanPath(outFn)

    def test_createEmptyImage(self):
        outFn = join('/tmp', 'empty.mrc')
        SIZE = (128, 128, 1, 1)
        ih = emlib.image.ImageHandler()
        DT = emlib.DT_FLOAT
        ih.createEmptyImage(outFn, SIZE[0], SIZE[1], dataType=DT)

        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), SIZE)
        self.assertEqual(ih.getDataType(outFn), DT)

    def test_scaleStack(self):
        particles = self.dataset.getFile("particles/BPV_1386.stk")
        outFn = join('/tmp', 'scaled.mrc')
        ih = emlib.image.ImageHandler()
        DT = emlib.DT_FLOAT

        # Scaled with a higher dimension using finalDimension parameter
        EXPECTED_SIZE = (256, 256, 131, 1)
        ih.scale2DStack(particles, outFn, finalDimension=EXPECTED_SIZE[0])

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), EXPECTED_SIZE)
        self.assertEqual(ih.getDataType(outFn), DT)

        # Scaled with a lower dimension using finalDimension parameter
        EXPECTED_SIZE = (64, 64, 131, 1)
        ih.scale2DStack(particles, outFn, finalDimension=EXPECTED_SIZE[0])

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), EXPECTED_SIZE)
        self.assertEqual(ih.getDataType(outFn), DT)

        # Scaled with a higher dimension using scaleFactor parameter
        scaleFactor = 1.5
        EXPECTED_SIZE = (210, 210, 131, 1)
        ih.scale2DStack(particles, outFn, scaleFactor=scaleFactor)

        self.assertTrue(os.path.exists(outFn))
        self.assertTrue(pwutils.getFileSize(outFn) > 0)
        self.assertEqual(ih.getDimensions(outFn), EXPECTED_SIZE)
        self.assertEqual(ih.getDataType(outFn), DT)


class TestSetOfMicrographs(BaseTest):
    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('xmipp_tutorial')
        cls.dbGold = cls.dataset.getFile('micsGoldSqlite')
        cls.micsPattern = cls.dataset.getFile('allMics')
        cls.dbFn = cls.getOutputPath('micrographs.sqlite')
        cls.acquisition = emobj.Acquisition(magnification=60000, voltage=300,
                                            sphericalAberration=2,
                                            amplitudeContrast=0.1)

        # cls.mics = glob(cls.micsPattern)
        cls.mics = []
        for mic in iglob(cls.micsPattern):
            cls.mics.append(cls.getRelPath(cls.dataset.getPath(), mic))

        if len(cls.mics) == 0:
            raise Exception('There are not micrographs matching pattern')
        cls.mics.sort()

    def checkSet(self, micSet):
        idCount = 1

        for mic1, fn in zip(micSet, self.mics):
            micFn = mic1.getFileName()
            self.assertEqual(fn, micFn,
                             "micrograph NAME in the set is wrong, \n   expected: '%s'\n        got: '%s'"
                             % (fn, micFn))
            self.assertEqual(idCount, mic1.getObjId(),
                             "micrograph ID in the set is wrong, \n   expected: '%s'\n        got: '%s'"
                             % (idCount, mic1.getObjId()))
            mic2 = micSet[idCount]  # Test getitem
            self.assertEqual(mic1.getObjId(), mic2.getObjId(), "micrograph got from ID is wrong")
            idCount += 1

    def testCreate(self):
        cwd = os.getcwd()
        # Change to test path
        os.chdir(self.dataset.getPath())
        """ Create a SetOfMicrographs from a list of micrographs """
        micSet = emobj.SetOfMicrographs(filename=self.dbFn)

        micSet.setAcquisition(self.acquisition)
        micSet.setSamplingRate(1.2)
        for fn in self.mics:
            mic = emobj.Micrograph()
            mic.setFileName(fn)
            micSet.append(mic)
        micSet.write()

        self.checkSet(micSet)

        # Check copy info also copies the samplingRate
        micSet2 = emobj.SetOfMicrographs(filename=':memory:')
        micSet2.copyInfo(micSet)
        self.assertAlmostEqual(micSet.getSamplingRate(), micSet2.getSamplingRate())

        os.chdir(cwd)

    def testRead(self):
        """ Read micrographs from a an sqlite file.
        It should contains Acquisition info. """
        micFn = self.dataset.getFile('micsGoldSqlite2')
        logger.info(">>> Reading gold micrographs from %s " % micFn)

        micSet = emobj.SetOfMicrographs(filename=micFn)
        self.assertEqual(2, micSet.getSize())
        acquisition = emobj.Acquisition()
        acquisition.setMagnification(10000.)
        acquisition.setVoltage(200.)
        acquisition.setSphericalAberration(2.26)
        acquisition.setAmplitudeContrast(0.1)

        mic2 = emobj.Micrograph()
        mic2.setSamplingRate(2.8)
        mic2.setAcquisition(acquisition)

        fileNames = ['/home/roberto/Scipion/Test/Test2/down2_12585',
                     '/home/roberto/Scipion/Test/Test2/down2_12610']
        counter = 0
        for mic in micSet:
            mic2.setFileName(fileNames[counter])
            self.assertTrue(mic.equalAttributes(mic2))
            counter += 1

    def test_mapper(self):
        """ test that indexes are created when a
        setOfParticles is created """
        MICNUMBER = 10
        indexesNames = ['_index']
        prot = createDummyProtocol("dummy_protocol")

        # create set of micrographs
        micSet = prot._createSetOfMicrographs()
        mic = emobj.Micrograph()
        for i in range(MICNUMBER + 1):
            mic.setLocation(i, "stack.stk")
            micSet.append(mic)
            mic.cleanObjId()
            micSet.write()

        # check defined indexes
        setOfMicrographsFileName = prot._getPath("micrographs.sqlite")
        indexes = sorted([index[1] for index in
                          getIndex(setOfMicrographsFileName)])
        for index, indexName in zip(indexes, indexesNames):
            self.assertEqual(index, 'index_' + indexName)


class TestSetOfParticles(BaseTest):
    """ Check if the information of the images is copied to another image when
    a new SetOfParticles is created"""

    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('xmipp_tutorial')

    def test_orderBy(self):
        # create setofProjections sorted
        imgSet1 = emobj.SetOfParticles(filename=':memory:', prefix='set1')
        imgSet2 = emobj.SetOfParticles(filename=':memory:', prefix='set2')
        imgSet1.setSamplingRate(1.5)
        imgSet2.setSamplingRate(1.5)
        img = emobj.Particle()

        for i in range(1, 10):
            img.setLocation(i, 'mystack.stk')
            img.setMicId(10 - i)
            img.setClassId(i % 5)
            imgSet1.append(img)
            img.setMicId(i)
            imgSet2.append(img)
            img.cleanObjId()
        # orderby
        for item1, item2 in zip(imgSet1.iterItems(orderBy='_micId',
                                                  direction='ASC'),
                                imgSet2.iterItems(orderBy='_micId',
                                                  direction='ASC')):
            self.assertEquals(item1.getMicId(), item2.getMicId())

    def test_readStack(self):
        """ Read an stack of 29 particles from .hdf file.
        Particles should be of 500x500 pixels.
        """
        size = 29
        xdim = 500
        inStack = self.dataset.getFile('particles1')
        outFn = self.getOutputPath('particles.sqlite')

        imgSet = emobj.SetOfParticles(filename=outFn)
        imgSet.setSamplingRate(1.0)
        imgSet.readStack(inStack)  # This should add 29 new items to the set

        self.assertEquals(size, imgSet.getSize())  # Check same size
        self.assertEquals(xdim, imgSet.getDim()[0])  # Check same dimensions

        logger.info("writing particles to %s" % outFn)
        imgSet.write()

        imgSet2 = emobj.SetOfParticles(filename=':memory:')
        imgSet2.copyInfo(imgSet)
        self.assertAlmostEqual(imgSet.getSamplingRate(), imgSet2.getSamplingRate())

    def test_hugeSet(self):
        """ Create a set of a big number of particles to measure
        creation time with sqlite operations. 
        """
        # Allow what huge means to be defined with environment var
        n = int(os.environ.get('SCIPION_TEST_HUGE', 10000))
        logger.info(">>>> Creating a set of %d particles." % n)
        logger.info("     (set SCIPION_TEST_HUGE environment var to other value)")

        dbFn = self.getOutputPath('huge_set.sqlite')
        # dbFn = ':memory:'

        img = emobj.Particle()
        imgSet = emobj.SetOfParticles(filename=dbFn)
        imgSet.setSamplingRate(1.0)

        for i in range(1, n + 1):
            # Creating object inside the loop significantly
            # decrease performance
            # img = Particle()
            img.setLocation(i, "images.stk")

            imgSet.append(img)
            img.cleanObjId()
            # img.setObjId(None)

        imgSet.write()

    def test_hugeSetToMd(self):
        """ Just as a benchrmark comparing to test_hugeSet ."""
        # Allow what huge means to be defined with environment var
        n = int(os.environ.get('SCIPION_TEST_HUGE', 10000))
        logger.info(">>>> Creating a set of %d particles." % n)
        logger.info("     (set SCIPION_TEST_HUGE environment var to other value)")

        imgMd = md.MetaData()

        for i in range(1, n + 1):
            # Creating object inside the loop significantly
            # decrease performance
            # img = Particle()
            objId = imgMd.addObject()
            imgMd.setValue(md.MDL_IMAGE, '%06d@images.stk' % (i + 1), objId)

        mdFn = self.getOutputPath('huge_set.xmd')
        logger.info("Writing metadata to %s." % mdFn)
        imgMd.write(mdFn)

    def test_hugeSetToText(self):
        """ Just as a benchrmark comparing to test_hugeSet ."""
        # Allow what huge means to be defined with environment var
        n = int(os.environ.get('SCIPION_TEST_HUGE', 10000))
        logger.info(">>>> Creating a set of %d particles." % n)
        logger.info("     (set SCIPION_TEST_HUGE environment var to other value)")

        textFn = self.getOutputPath('huge_set.txt')
        logger.info("Writing to text file %s " % textFn)
        f = open(textFn, 'w')

        for i in range(1, n + 1):
            string = '%06d@images.stk' % i
            f.write(string)

        f.close()

    def test_getFiles(self):
        # create setofImages
        dbFn = self.getOutputPath('multistack_set.sqlite')
        # dbFn = ':memory:'
        n = 10
        m = 3
        img = emobj.Particle()
        imgSet = emobj.SetOfParticles(filename=dbFn)
        imgSet.setSamplingRate(1.0)
        goldStacks = set()

        for j in range(1, m + 1):
            stackName = 'stack%02d.stk' % j
            goldStacks.add(stackName)
            for i in range(1, n + 1):
                img.setLocation(i, stackName)
                imgSet.append(img)
                img.cleanObjId()

        self.assertEquals(goldStacks, imgSet.getFiles())

        imgSet.close()

        # It should automatically call load
        # before accessing items db        
        imgSet.getFirstItem()

    def test_mapper(self):
        """ test that indexes are created when a
        setOfParticles is created """
        PARTNUMBER = 10
        indexesNames = ['_classId', '_micId']
        prot = createDummyProtocol("dummy_protocol")

        # create set of particles
        partSet = prot._createSetOfParticles()
        part = emobj.Particle()
        for i in range(PARTNUMBER + 1):
            part.setLocation(i, "stack.vol")
            partSet.append(part)
            part.cleanObjId()
        partSet.write()

        # check defined indexes
        setOfPArticleFileName = prot._getPath("particles.sqlite")
        indexes = sorted([index[1] for index in
                          getIndex(setOfPArticleFileName)])
        for index, indexName in zip(indexes, indexesNames):
            self.assertEqual(index, 'index_' + indexName)


class TestSetOfCoordinates(BaseTest):
    # TODO: A proper test for setOfCoordinates is missing
    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)

    def test_mapper(self):
        """ test that indexes are created when a
        setOfCoordinates is created """
        PARTNUMBER = 10
        MICNUMBER = 600
        NUMBERCOORDINATES = PARTNUMBER * MICNUMBER
        indexesNames = ['_micId']
        prot = createDummyProtocol("dummy_protocol")

        # create set of micrographs
        micSet = emobj.SetOfMicrographs(filename=":memory:")
        mic = emobj.Micrograph()
        for i in range(NUMBERCOORDINATES):
            mic.setLocation(i, "mic_%06d.mrc" % i)
            micSet.append(mic)
            mic.cleanObjId()
        micSet.write()

        # create a set of particles
        coordSet = prot._createSetOfCoordinates(micSet)
        coord = emobj.Coordinate()
        for i in range(NUMBERCOORDINATES):
            coordSet.append(coord)
            coord.cleanObjId()
        coordSet.write()

        # check defined indexes
        setOfCoordinatesFileName = \
            prot._getPath("coordinates.sqlite")
        logger.info(os.path.abspath(setOfCoordinatesFileName))
        indexes = sorted([index[1] for index in
                          getIndex(setOfCoordinatesFileName)])
        for index, indexName in zip(indexes, indexesNames):
            self.assertEqual(index, 'index_' + indexName)

        # Test speed: based on loop in file protocol_extractparticles.py
        # for 600 mic and 100 part the values for the first
        # second and third  case where:
        # Loop with index: 5 sec
        # Loop no index: 8:01 min
        # Loop optimized code: 4 sec
        # for 6000 mic and 200 part the values for the first
        # Loop with index: 1:47 min
        # optimized Loop with index: 1:20 min
        # Loop no index: after several  hours I stopped the process

        SPEEDTEST = True
        if SPEEDTEST:  # code from protocol_particles. line 415
            testTimer = pwutils.Timer()
            testTimer.tic()
            for mic in micSet:
                micId = mic.getObjId()
                coordList = []
                for coord in coordSet.iterItems(where='_micId=%s' % micId):
                    coordList.append(coord.clone())
            testTimer.toc("Loop with INDEX took:")

            lastMicId = None
            testTimer.tic()
            for coord in coordSet.iterItems(orderBy='_micId',
                                            direction='ASC'):
                micId = coord.getMicId()
                if micId != lastMicId:
                    lastMicId = micId
                    coordList = []
                coordList.append(coord.clone())
            testTimer.toc("Loop with INDEX and proper code, took:")

            # delete INDEX, this will not work
            # if database is not sqlite
            conn = sqlite3.connect(setOfCoordinatesFileName)
            cur = conn.cursor()
            for index in indexesNames:
                cur.execute("DROP INDEX index_%s" % index)
            cur.close()
            conn.close()

            testTimer.tic()
            for mic in micSet:
                micId = mic.getObjId()
                coordList = []
                for coord in coordSet.iterItems(where='_micId=%s' % micId):
                    coordList.append(coord.clone())
            testTimer.toc("Loop with NO INDEX took:")

            lastMicId = None
            testTimer.tic()
            for coord in coordSet.iterItems(orderBy='_micId',
                                            direction='ASC'):
                micId = coord.getMicId()
                if micId != lastMicId:
                    lastMicId = micId
                    coordList = []
                coordList.append(coord.clone())
            testTimer.toc("Loop with NO INDEX but proper code, took:")


class TestSetOfClasses2D(BaseTest):
    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('model')
        cls.selectionFn = cls.dataset.getFile('classesSelection')

    def test_basics(self):
        """ Load an existing SetOfClasses and test basic properties 
        such us: _mapperPath, iteration and others.
        """
        classes2DSet = emobj.SetOfClasses2D(filename=self.selectionFn)
        # Check the _mapperPath was properly set
        self.assertEqual(classes2DSet._mapperPath.get(), '%s, ' % self.selectionFn)

        cls1 = classes2DSet.getFirstItem()
        self.assertEqual(cls1._mapperPath.get(), '%s,Class001' % self.selectionFn)

        img1 = cls1.getFirstItem()
        self.assertEqual(img1.getObjId(), 1)  # First image of first class is 1 in this test

        images = [[1, 3, 4, 5, 7, 9, 11, 14, 16, 17, 21, 22, 24, 26, 28, 29, 30,
                   35, 37, 38, 39, 40, 42, 45, 54, 57, 62, 65, 67, 69, 70, 71, 74],
                  [2, 8, 10, 12, 13, 15, 19, 20, 23, 25, 27, 33, 34, 41, 43, 44,
                   46, 47, 48, 49, 50, 51, 52, 53, 55, 56, 58, 59, 60, 61, 63,
                   64, 68, 72, 73, 75, 76],
                  [18, 31, 32],
                  [6, 36, 66]]

        classes2DSet.close()
        # Check iteration is working properly even after 
        # a close operation, it should open automatically
        for i, cls in enumerate(classes2DSet):
            l = images[i]
            for j, img in enumerate(cls):
                self.assertEquals(img.getObjId(), l[j])

        for i, rep in enumerate(classes2DSet.iterRepresentatives()):
            self.assertIsNotNone(rep.getLocation())

        # Check the SetOfClasses.iterClassItems method
        allImages = [img for imgList in images for img in imgList]
        idsImages = [img.getObjId()
                     for img in classes2DSet.iterClassItems(iterDisabled=True)]
        self.assertEqual(allImages, idsImages)

    def test_subsetsFromSelection(self):
        """ From a sqlite file of a SetOfClasses2D, with some
        classes and element disabled, we want to create a 
        subset of images and classes.
        """
        classes2DSet = emobj.SetOfClasses2D(filename=self.selectionFn)

        imgSet = emobj.SetOfParticles(filename=':memory:')
        # We are going to iterate over the enabled items and create
        # a new set of images
        imgSet.appendFromClasses(classes2DSet)
        # Since we have disabled two classes (6 images) and 1 images
        # from class 1 and 2 (2 images), the final size of the set
        # should be 68
        sizes = [32, 36]
        self.assertEqual(imgSet.getSize(), sum(sizes))
        imgSet.clear()  # Close db connection and clean data

        # Now create a subset of classes and check the number
        # of images per class
        clsSet = emobj.SetOfClasses2D(filename=':memory:')
        clsSet.appendFromClasses(classes2DSet)
        for i, cls in enumerate(clsSet):
            self.assertEqual(cls.getSize(), sizes[i])
        clsSet.clear()  # Close db connection and clean data


class TestTransform(BaseTest):

    def test_scale(self):
        """ Check Scale storage in transformation class
        """
        t = emobj.Transform()
        m = t.getMatrix()
        m[0, 3] = 2
        m[1, 3] = 4
        m[2, 3] = 6
        m[3, 3] = 5
        t.scale(0.5)

        self.assertAlmostEqual(m[0, 3], 1)
        self.assertAlmostEqual(m[1, 3], 2)
        self.assertAlmostEqual(m[2, 3], 3)
        self.assertAlmostEqual(m[3, 3], 1)

    def test_scaleShifts(self):
        """ Check Scale 2D shifts in transformation class
        """
        t = emobj.Transform()
        m = t.getMatrix()
        m[0, 3] = 2
        m[1, 3] = 4
        m[2, 3] = 6
        m[3, 3] = 5
        t.scaleShifts(0.5)

        self.assertAlmostEqual(m[0, 3], 1)
        self.assertAlmostEqual(m[1, 3], 2)
        self.assertAlmostEqual(m[2, 3], 3)
        self.assertAlmostEqual(m[3, 3], 5)

    def test_clone(self):
        """ Check that cloning the Transform will 
        also copy the values of underlying numpy matrix.
        """
        t = emobj.Transform()
        m = t.getMatrix()
        m[0, 3] = 2
        m[1, 3] = 4

        t2 = t.clone()
        m2 = t2.getMatrix()
        self.assertTrue(np.allclose(m, m2, rtol=1e-2))

        p = emobj.Particle()
        p.setTransform(t)

        p2 = p.clone()
        m3 = p2.getTransform().getMatrix()
        self.assertTrue(np.allclose(m, m3, rtol=1e-2))


class TestCopyItems(BaseTest):
    _labels = [SMALL, WEEKLY]

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        cls.dataset = DataSet.getDataSet('model')

    def test_listAttribute(self):
        """ 
        Test the use of copyItems function where the items have
        a property that is a list.
        """

        inFn = self.dataset.getFile('particles/particles_nma.sqlite')
        outFn = self.getOutputPath('particles.sqlite')

        inputSet = emobj.SetOfParticles(filename=inFn)
        inputSet.setSamplingRate(1.0)

        outputSet = emobj.SetOfParticles(filename=outFn)
        outputSet.copyInfo(inputSet)

        outputSet.copyItems(inputSet,
                            updateItemCallback=self._updateItem)

        outputSet.write()
        outputSet.close()

        # Read again the written set of particles
        checkSet = emobj.SetOfParticles(filename=outFn)

        # check that the first item (as the rest)
        # have the added _list attribute with the correct values
        particle = checkSet.getFirstItem()
        self.assertTrue(particle.hasAttribute('_list'))
        for v1, v2 in zip([1.0, 2.0], particle._list):
            self.assertAlmostEqual(v1, float(v2))

        # check that copied items have exactly
        # the same attributes than the input set
        # except for the _list
        for i1, i2 in zip(inputSet, checkSet):
            self.assertTrue(i2.equalAttributes(i1, ignore=['_list']))

    def _updateItem(self, item, row):
        item._list = emobj.CsvList()
        item._list.set([1.0, 2.0])


class TestCoordinatesTiltPair(BaseTest):
    # TODO: A proper test for CoordinatesTiltPair is missing
    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)

    def test_mapper(self):
        """ test that indexes are created when a
        setOfCoordinates is created """
        MICNUMBER = 10
        indexesNames = ['_untilted._micId',
                        '_tilted._micId']
        prot = createDummyProtocol("dummy_protocol")

        # create set of untilted and tilted micrographs
        uMicSet = emobj.SetOfMicrographs(filename=":memory:")
        tMicSet = emobj.SetOfMicrographs(filename=":memory:")
        mic = emobj.Micrograph()
        for i in range(MICNUMBER):
            mic.setLocation(i, "umic_%06d.mrc" % i)
            uMicSet.append(mic)
            mic.cleanObjId()
            mic.setLocation(i, "tmic_%06d.mrc" % i)
            tMicSet.append(mic)
            mic.cleanObjId()
        # TODO I do not see any example on how to create
        # a set of CoordinatesTiltPair. I can image
        # that a need to create a set of micTiltPairs
        # and two sets of coordinates but the person who
        # added that data type Should provide a clear test
        # when this is done then I will finish the test_mapper


class TestCTFModel(TestCase):
    def test_stringContext(self):
        ctf = emobj.CTFModel()
        # All values to None should be printable without exception
        logger.info("Empty Ctf str converts to %s" % ctf)

        ctf.setDefocusU(2.4545)
        logger.info("Ctf with defocusU converts to %s" % ctf)
        self.assertTrue("2.45," in str(ctf))


class TestSequenceHandler(pwtests.BaseTest):
    NAME = 'USER_SEQ'
    DESCRIPTION = 'User description'
    UNIPROTID1 = 'P12345'
    UNIPROTID2 = 'P03252'

    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)

    def testImportStructureAminoacidSequence1(self):
        """
        Import the sequence of chain B of atomic structure 3lqd.cif
        """
        args = {'inputSequenceName': self.NAME,
                'inputProteinSequence':
                    emprot.ProtImportSequence.IMPORT_FROM_UNIPROT,
                'uniProtSequence': self.UNIPROTID1,
                }
        prot1 = self.newProtocol(emprot.ProtImportSequence, **args)
        self.launchProtocol(prot1)
        sequence1 = prot1.outputSequence

        args = {'inputSequenceName': self.NAME,
                'inputProteinSequence':
                    emprot.ProtImportSequence.IMPORT_FROM_UNIPROT,
                'uniProtSequence': self.UNIPROTID2,
                }
        prot2 = self.newProtocol(emprot.ProtImportSequence, **args)
        self.launchProtocol(prot2)
        sequence2 = prot2.outputSequence

        #Saving Sequences into files
        outFile = prot1._getPath('sequence.fasta')
        sequence1.exportToFile(outFile)
        # sequence2.exportToFile(prot2._getPath('sequence.genbank'))

        #Reading exported sequence
        readSequence = SequenceHandler()
        sequencesDics = readSequence.readSequencesFromFile(outFile)
        self.assertEqual(sequence1.getSequence(), sequencesDics[0]['sequence'])

        #Appending sequence to existing sequence file
        sequence2.appendToFile(outFile)

        # Reading appended sequences
        readSequence = SequenceHandler()
        sequencesDics = readSequence.readSequencesFromFile(outFile)
        self.assertEqual(sequence1.getSequence(), sequencesDics[0]['sequence'])
        self.assertEqual(sequence2.getSequence(), sequencesDics[1]['sequence'])


